/*
 * $Id: BreakpointHitListenerAdapter.java,v 1.4 2005/08/28 11:29:59 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 04.01.2005 15:41:31 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher;

import java.util.List;
import java.util.Vector;

import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.core.dom.Message;
import org.eclipse.jdt.debug.core.IJavaBreakpoint;
import org.eclipse.jdt.debug.core.IJavaBreakpointListener;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaLineBreakpoint;
import org.eclipse.jdt.debug.core.IJavaMethodEntryBreakpoint;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaType;

/**
 * Use the breakpoint hit listener adapter as an extension of the
 * IJavaBreakpointListener if you want listen on the breakpoint hit event. This
 * adapter also provides functionality to focus on one or more breakpoints given
 * to him.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
 */
public abstract class BreakpointHitListenerAdapter implements
		IJavaBreakpointListener {

	/** Personal logger for this class */
	private final static DDLogger logger = DDStateLaunchPlugin
			.getLogger(BreakpointHitListenerAdapter.class);

	/** A list of all breakpoints which this listener should listened */
	private List fBreakpoints = null;

	/**
	 * Remember the current java thread if a breakpoint in the breakpoint list
	 * was hitten
	 */
	private IJavaThread fCurrentJavaThread = null;

	/**
	 * Flag if this listener is enabled or not. If it is disabled and a
	 * breakpoint of the breakpoint list will hitten this listener return the
	 * command, which is specified in <code>fDisableCommand</code>.
	 */
	private boolean fEnabled = true;

	/**
	 * The command that the <code>breakpointHit</code> function will return if
	 * this listener is disabled and one or more of the listeners associated
	 * breakpoints are hitted.
	 */
	private int fDisableCommand = IJavaBreakpointListener.DONT_CARE;

	/**
	 * Use this field to remember if an error occured.
	 */
	private Throwable fError = null;

	public BreakpointHitListenerAdapter(IJavaBreakpoint bp) {
		this.fBreakpoints = new Vector();
		this.fBreakpoints.add(bp);
		this.fEnabled = true;
	}

	public BreakpointHitListenerAdapter(IJavaBreakpoint[] breakpoints) {
		this.fBreakpoints = new Vector();

		for (int i = 0; i < breakpoints.length; i++) {
			IJavaBreakpoint breakpoint = breakpoints[i];
			this.fBreakpoints.add(breakpoint);
		}

		this.fEnabled = true;
	}

	/**
	 * Returns the current thread if the a breakpoint was hit and the thread was
	 * suspended
	 * 
	 * @return the current thread if the a breakpoint was hit and the thread was
	 *         suspended
	 */
	public IJavaThread getCurrentThread() {
		return this.fCurrentJavaThread;
	}

	/**
	 * Returns true if an error occured, otherwise false.
	 * 
	 * @return true if an error occured, otherwise false.
	 */
	public boolean hasErros() {
		return this.fError != null;
	}

	/**
	 * Returns the exception object if an error occured, otherwise returns
	 * <code>null</code>.
	 * 
	 * @return the exception object if an error occured, otherwise returns
	 *         <code>null</code>.
	 */
	public Throwable getError() {
		return this.fError;
	}

	/**
	 * Set the error object. If you want to reset the error give
	 * <code>null</code> as parameter.
	 * 
	 * @param t
	 *            the error object or <code>null</code>.
	 */
	public void setError(Throwable t) {
		this.fError = t;
	}

	/**
	 * Enable this breakpoint listener.
	 */
	public void enable() {
		this.fEnabled = true;
	}

	/**
	 * Disable this breakpoint listener. Return the given command if one of the
	 * breakpoints associated with this listener will hitted.
	 * 
	 * @param command
	 *            whether the thread should suspend or whether this listener
	 *            doesn't care - one of <code>SUSPEND</code>,
	 *            <code>DONT_SUSPEND</code>, or <code>DONT_CARE</code>
	 */
	public void disable(int command) {

		if ((command & (SUSPEND | DONT_CARE | DONT_SUSPEND)) == 0) {
			throw new IllegalArgumentException(
					"The command should be either SUSPEND, DONT_SUSPEND or DONT_CARE");
		}

		this.fDisableCommand = command;
		this.fEnabled = false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#addingBreakpoint(org.eclipse.jdt.debug.core.IJavaDebugTarget,
	 *      org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public void addingBreakpoint(IJavaDebugTarget target,
			IJavaBreakpoint breakpoint) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#installingBreakpoint(org.eclipse.jdt.debug.core.IJavaDebugTarget,
	 *      org.eclipse.jdt.debug.core.IJavaBreakpoint,
	 *      org.eclipse.jdt.debug.core.IJavaType)
	 */
	public int installingBreakpoint(IJavaDebugTarget target,
			IJavaBreakpoint breakpoint, IJavaType type) {
		int result = DONT_CARE;

		if (this.fBreakpoints.contains(breakpoint) && this.fEnabled) {
			result = INSTALL;
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#breakpointInstalled(org.eclipse.jdt.debug.core.IJavaDebugTarget,
	 *      org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public void breakpointInstalled(IJavaDebugTarget target,
			IJavaBreakpoint breakpoint) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#breakpointHit(org.eclipse.jdt.debug.core.IJavaThread,
	 *      org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public int breakpointHit(IJavaThread thread, IJavaBreakpoint breakpoint) {

		if (!this.fBreakpoints.contains(breakpoint)) {
			return IJavaBreakpointListener.DONT_CARE;
		}

		if (!this.fEnabled) {
			return this.fDisableCommand;
		}

		// remember the current java thread
		this.fCurrentJavaThread = thread;

		this.debugOutput(breakpoint);

		int result = this.givenBreakpointHit(thread, breakpoint);

		return result;
	}

	/**
	 * Trace some informations about the given breakpoint.
	 * 
	 * @param breakpoint
	 */
	private void debugOutput(IJavaBreakpoint breakpoint) {
		String typeName = "n.a.";
		try {
			typeName = breakpoint.getTypeName();
		} catch (CoreException e) {
			// do nothing
		}

		if (breakpoint instanceof IJavaMethodEntryBreakpoint) {
			String methodName = "n.a.";
			String methodSignature = "n.a";

			try {
				methodName = ((IJavaMethodEntryBreakpoint) breakpoint)
						.getMethodName();
			} catch (CoreException e) {
				// do nothing
			}

			try {
				methodSignature = ((IJavaMethodEntryBreakpoint) breakpoint)
						.getMethodSignature();
			} catch (CoreException e) {
				// do nothing
			}

			logger.trace(DDStateLaunchPlugin.DEBUG, "Method entry breakpoint "
					+ typeName + "(" + methodName + " - " + methodSignature
					+ ") reached");

		} else if (breakpoint instanceof IJavaLineBreakpoint) {
			String lineNumber = "n.a.";
			try {
				lineNumber = ""
						+ ((IJavaLineBreakpoint) breakpoint).getLineNumber();
			} catch (CoreException e) {
				// do nothing
			}

			logger.trace(DDStateLaunchPlugin.DEBUG, "Line breakpoint "
					+ typeName + "(" + lineNumber + ") reached");
		}
	}

	/**
	 * This method will be called if one of the given breakpoints is hit.
	 * 
	 * Allows this listener to vote to determine if the given thread should be
	 * suspended in reponse to the breakpoint. If at least one listener votes to
	 * SUSPEND, the thread will suspend. If there are no votes to suspend the
	 * thread, there must be at least one DONT_SUSPEND vote to avoid the
	 * suspension (resume). If all listeners vote DONT_CARE, the thread will
	 * suspend by default.
	 * 
	 * @param thread
	 *            Java thread
	 * @param breakpoint
	 *            Java breakpoint
	 * @return whether the thread should suspend or whether this listener
	 *         doesn't care - one of SUSPEND, DONT_SUSPEND, or DONT_CARE
	 */
	protected abstract int givenBreakpointHit(IJavaThread thread,
			IJavaBreakpoint breakpoint);

	/**
	 * If the program execution stop on one of the given breakpoints of this
	 * listener you can terminate the java thread with this mehtod.
	 * 
	 * @throws DebugException
	 *             on failure. Reasons include:
	 *             <ul>
	 *             <li>TARGET_REQUEST_FAILED - The request failed in the target
	 *             <li>NOT_SUPPORTED - The capability is not supported by the
	 *             target
	 *             </ul>
	 */
	public void terminateCurrentThread() throws DebugException {
		if (this.fCurrentJavaThread != null) {
			if (this.fCurrentJavaThread.canTerminate()) {
				this.fCurrentJavaThread.terminate();
			}
		}
	}

	/**
	 * If the program execution stop on one of the given breakpoints of this
	 * listener you can resume the java thread with this mehtod.
	 * 
	 * @throws DebugException
	 *             on failure. Reasons include:
	 *             <ul>
	 *             <li>TARGET_REQUEST_FAILED - The request failed in the target
	 *             <li>NOT_SUPPORTED - The capability is not supported by the
	 *             target
	 *             </ul>
	 */
	public void resumeCurrentThread() throws DebugException {
		if (this.fCurrentJavaThread != null) {
			if (this.fCurrentJavaThread.canResume()) {
				this.fCurrentJavaThread.resume();
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#breakpointRemoved(org.eclipse.jdt.debug.core.IJavaDebugTarget,
	 *      org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public void breakpointRemoved(IJavaDebugTarget target,
			IJavaBreakpoint breakpoint) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#breakpointHasRuntimeException(org.eclipse.jdt.debug.core.IJavaLineBreakpoint,
	 *      org.eclipse.debug.core.DebugException)
	 */
	public void breakpointHasRuntimeException(IJavaLineBreakpoint breakpoint,
			DebugException exception) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.debug.core.IJavaBreakpointListener#breakpointHasCompilationErrors(org.eclipse.jdt.debug.core.IJavaLineBreakpoint,
	 *      org.eclipse.jdt.core.dom.Message[])
	 */
	public void breakpointHasCompilationErrors(IJavaLineBreakpoint breakpoint,
			Message[] errors) {
		// do nothing
	}
}