/*
 * $Id: BreakpointManager.java,v 1.4 2005/08/21 13:14:05 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Feb 10, 2005 11:01:11 AM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher;

import java.util.Vector;

import org.deltadebugging.ddstate.launcher.matcher.breakpoint.IBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.ILineBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.IMethodEntryBreakpointLocation;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IMethodIdentifier;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.IBreakpointListener;
import org.eclipse.debug.core.IBreakpointManager;
import org.eclipse.debug.core.IBreakpointManagerListener;
import org.eclipse.debug.core.IBreakpointsListener;
import org.eclipse.debug.core.model.IBreakpoint;
import org.eclipse.jdt.debug.core.IJavaBreakpoint;
import org.eclipse.jdt.debug.core.IJavaLineBreakpoint;
import org.eclipse.jdt.debug.core.IJavaMethodEntryBreakpoint;
import org.eclipse.jdt.debug.core.JDIDebugModel;

/**
 * COMMENTME BreakpointManager
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class BreakpointManager implements IExtBreakpointManager {

	// Personal logger for this class
	final static DDLogger logger = DDStateLaunchPlugin
			.getLogger(BreakpointManager.class);

	BreakpointManager() {
		// util class is not initializable
	}

	/**
	 * Returns the underlying breakpoint manager for this extended manager. The
	 * underlying manager provide the core functionality.
	 * 
	 * @return the underlying breakpoint manager, which provide the core
	 *         functionality to manage breakpoints and breakpoint listeners.
	 */
	private IBreakpointManager getUnderlyingBPManager() {
		return DebugPlugin.getDefault().getBreakpointManager();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#deleteExistingBreakpoint(org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public void deleteExistingBreakpoint(IJavaBreakpoint breakpoint)
			throws CoreException {
		if (breakpoint != null) {
			if (getUnderlyingBPManager().isRegistered(breakpoint)) {
				this.unregisterBreakpoint(breakpoint, true);
			} else {
				breakpoint.delete();
			}

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#deleteExistingBreakpoints(org.eclipse.jdt.debug.core.IJavaBreakpoint[])
	 */
	public void deleteExistingBreakpoints(IJavaBreakpoint[] breakpoints)
			throws CoreException {
		for (int i = 0; i < breakpoints.length; i++) {
			IJavaBreakpoint breakpoint = breakpoints[i];

			this.deleteExistingBreakpoint(breakpoint);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#registerBreakpoint(org.eclipse.jdt.debug.core.IJavaBreakpoint)
	 */
	public void registerBreakpoint(IJavaBreakpoint breakpoint)
			throws CoreException {
		IBreakpointManager breakpointManager = getUnderlyingBPManager();

		if (!breakpointManager.isRegistered(breakpoint)) {
			breakpointManager.addBreakpoint(breakpoint);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#unregisterBreakpoint(org.eclipse.jdt.debug.core.IJavaBreakpoint,
	 *      boolean)
	 */
	public void unregisterBreakpoint(IJavaBreakpoint breakpoint, boolean delete)
			throws CoreException {
		IBreakpointManager breakpointManager = getUnderlyingBPManager();

		if (breakpointManager.isRegistered(breakpoint)) {
			breakpointManager.removeBreakpoint(breakpoint, delete);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#disableAllExistingBreakpoints()
	 */
	public IJavaBreakpoint[] disableAllExistingBreakpoints()
			throws CoreException {
		Vector disabledBreakpoints = new Vector();

		IBreakpoint[] breakpoints = this.getUnderlyingBPManager()
				.getBreakpoints();

		for (int i = 0; i < breakpoints.length; i++) {
			IBreakpoint breakpoint = breakpoints[i];

			if (breakpoint.isEnabled()) {
				breakpoint.setEnabled(false);
				disabledBreakpoints.add(breakpoint);
			}
		}

		return (IJavaBreakpoint[]) disabledBreakpoints
				.toArray(new IJavaBreakpoint[disabledBreakpoints.size()]);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#enabledBreakpoints(org.eclipse.jdt.debug.core.IJavaBreakpoint[])
	 */
	public void enabledBreakpoints(IJavaBreakpoint[] breakpoints)
			throws CoreException {

		for (int i = 0; i < breakpoints.length; i++) {
			IJavaBreakpoint breakpoint = breakpoints[i];

			if (breakpoint.isRegistered()) {
				breakpoint.setEnabled(true);
			}

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#createLineBreakPoint(java.lang.String,
	 *      int, int, boolean)
	 */
	public IJavaLineBreakpoint createLineBreakPoint(String fullName, int line,
			int hitCount, boolean register) throws CoreException {
		IJavaLineBreakpoint result = null;

		result = JDIDebugModel.createLineBreakpoint(ResourcesPlugin
				.getWorkspace().getRoot(), fullName, line, -1, -1, hitCount,
				register, null);

		// if the breakpoint is disabled enabled it
		if (!result.isEnabled()) {
			result.setEnabled(true);
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#createMethodEntryBreakpoint(org.deltadebugging.ddstate.memorygraph.IMethodIdentifier,
	 *      int, boolean)
	 */
	public IJavaMethodEntryBreakpoint createMethodEntryBreakpoint(
			IMethodIdentifier methodIdentifier, int hitCount, boolean register)
			throws CoreException {
		IJavaMethodEntryBreakpoint result = null;

		result = JDIDebugModel.createMethodEntryBreakpoint(ResourcesPlugin
				.getWorkspace().getRoot(), methodIdentifier.getTypeName(),
				methodIdentifier.getMethodName(), methodIdentifier
						.getSignature(), -1, -1, -1, hitCount, register, null);

		// if the breakpoint is disabled enabled it
		if (!result.isEnabled()) {
			result.setEnabled(true);
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#addBreakpoint(org.eclipse.debug.core.model.IBreakpoint)
	 */
	public void addBreakpoint(IBreakpoint breakpoint) throws CoreException {
		this.getUnderlyingBPManager().addBreakpoint(breakpoint);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#addBreakpoints(org.eclipse.debug.core.model.IBreakpoint[])
	 */
	public void addBreakpoints(IBreakpoint[] breakpoints) throws CoreException {
		this.getUnderlyingBPManager().addBreakpoints(breakpoints);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#getBreakpoint(org.eclipse.core.resources.IMarker)
	 */
	public IBreakpoint getBreakpoint(IMarker marker) {
		return this.getUnderlyingBPManager().getBreakpoint(marker);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#getBreakpoints()
	 */
	public IBreakpoint[] getBreakpoints() {
		return this.getUnderlyingBPManager().getBreakpoints();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#hasBreakpoints()
	 */
	public boolean hasBreakpoints() {
		return this.getUnderlyingBPManager().hasBreakpoints();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#getBreakpoints(java.lang.String)
	 */
	public IBreakpoint[] getBreakpoints(String modelIdentifier) {
		return this.getUnderlyingBPManager().getBreakpoints(modelIdentifier);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#isRegistered(org.eclipse.debug.core.model.IBreakpoint)
	 */
	public boolean isRegistered(IBreakpoint breakpoint) {
		return this.getUnderlyingBPManager().isRegistered(breakpoint);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#fireBreakpointChanged(org.eclipse.debug.core.model.IBreakpoint)
	 */
	public void fireBreakpointChanged(IBreakpoint breakpoint) {
		this.fireBreakpointChanged(breakpoint);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#removeBreakpoint(org.eclipse.debug.core.model.IBreakpoint,
	 *      boolean)
	 */
	public void removeBreakpoint(IBreakpoint breakpoint, boolean delete)
			throws CoreException {
		this.getUnderlyingBPManager().removeBreakpoint(breakpoint, delete);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#removeBreakpoints(org.eclipse.debug.core.model.IBreakpoint[],
	 *      boolean)
	 */
	public void removeBreakpoints(IBreakpoint[] breakpoints, boolean delete)
			throws CoreException {
		this.getUnderlyingBPManager().removeBreakpoints(breakpoints, delete);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#addBreakpointListener(org.eclipse.debug.core.IBreakpointListener)
	 */
	public void addBreakpointListener(IBreakpointListener listener) {
		this.getUnderlyingBPManager().addBreakpointListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#removeBreakpointListener(org.eclipse.debug.core.IBreakpointListener)
	 */
	public void removeBreakpointListener(IBreakpointListener listener) {
		this.getUnderlyingBPManager().removeBreakpointListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#addBreakpointListener(org.eclipse.debug.core.IBreakpointsListener)
	 */
	public void addBreakpointListener(IBreakpointsListener listener) {
		this.getUnderlyingBPManager().addBreakpointListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#removeBreakpointListener(org.eclipse.debug.core.IBreakpointsListener)
	 */
	public void removeBreakpointListener(IBreakpointsListener listener) {
		this.getUnderlyingBPManager().removeBreakpointListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#addBreakpointManagerListener(org.eclipse.debug.core.IBreakpointManagerListener)
	 */
	public void addBreakpointManagerListener(IBreakpointManagerListener listener) {
		this.getUnderlyingBPManager().addBreakpointManagerListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#removeBreakpointManagerListener(org.eclipse.debug.core.IBreakpointManagerListener)
	 */
	public void removeBreakpointManagerListener(
			IBreakpointManagerListener listener) {
		this.getUnderlyingBPManager().removeBreakpointManagerListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#isEnabled()
	 */
	public boolean isEnabled() {
		return this.getUnderlyingBPManager().isEnabled();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled) {
		this.getUnderlyingBPManager().setEnabled(enabled);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.IBreakpointManager#getTypeName(org.eclipse.debug.core.model.IBreakpoint)
	 */
	public String getTypeName(IBreakpoint breakpoint) {
		return this.getUnderlyingBPManager().getTypeName(breakpoint);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.IExtBreakpointManager#createBreakpoint(org.deltadebugging.ddstate.launcher.matcher.breakpoint.IBreakpointDesc,
	 *      boolean)
	 */
	public IJavaBreakpoint createBreakpoint(
			IBreakpointLocation breakpointLocation, boolean register)
			throws CoreException {

		IJavaBreakpoint result = null;

		if (breakpointLocation instanceof ILineBreakpointLocation) {
			ILineBreakpointLocation lineBPDesc = (ILineBreakpointLocation) breakpointLocation;

			result = this.createLineBreakPoint(lineBPDesc.getTypeName(),
					lineBPDesc.getLineNumber(), lineBPDesc.getHitCount(),
					register);

		} else if (breakpointLocation instanceof IMethodEntryBreakpointLocation) {
			IMethodEntryBreakpointLocation methodBPDesc = (IMethodEntryBreakpointLocation) breakpointLocation;

			result = this.createMethodEntryBreakpoint(methodBPDesc
					.getMethodIdentifier(), methodBPDesc.getHitCount(),
					register);

		}

		return result;
	}
}