/*
 * $Id: DDStateBaseLaunchConfiguration.java,v 1.30 2005/10/25 10:57:58 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 04.01.2005 11:39:33 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher;

import org.deltadebugging.ddstate.launcher.client.IDDLauncher;
import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;
import org.deltadebugging.ddstate.launcher.client.cts.CTSLauncher;
import org.deltadebugging.ddstate.launcher.client.testrunner.JUnitTestDescription;
import org.deltadebugging.ddstate.launcher.ddrun.DDLaunchDescription;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.matcher.IMatchPoint;
import org.deltadebugging.ddstate.launcher.matcher.MatchPoint;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.IBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.LineBreakpointLocation;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.ui.DebugUITools;
import org.eclipse.debug.ui.ILaunchGroup;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.launching.AbstractJavaLaunchConfigurationDelegate;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;

/**
 * COMMENTME Comment type DDStateBaseLaunchConfiguration
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DDStateBaseLaunchConfiguration extends
        AbstractJavaLaunchConfigurationDelegate {

    // Personal logger for this class
    final static DDLogger logger = DDStateLaunchPlugin
            .getLogger(DDStateBaseLaunchConfiguration.class);

    /*
     * (non-Javadoc)
     * 
     * @see org.eclipse.debug.core.model.ILaunchConfigurationDelegate#launch(org.eclipse.debug.core.ILaunchConfiguration,
     *      java.lang.String, org.eclipse.debug.core.ILaunch,
     *      org.eclipse.core.runtime.IProgressMonitor)
     */
    public void launch(ILaunchConfiguration configuration, String mode,
            ILaunch launch, IProgressMonitor monitor) throws CoreException {

        // read required variables from the configuration
        String passTestName = configuration.getAttribute(
                IDDLaunchConstants.PASSING_TEST_NAME, "");

        String passBPClassName = configuration.getAttribute(
                IDDLaunchConstants.PASSING_BP_CLASS_NAME, "");
        int passBPLineNumber = configuration.getAttribute(
                IDDLaunchConstants.PASSING_BP_LINE_NUMBER, -1);
        int passBPHitCount = configuration.getAttribute(
                IDDLaunchConstants.PASSING_BP_HIT_COUNT, 0);

        String failTestName = configuration.getAttribute(
                IDDLaunchConstants.FAILING_TEST_NAME, "");

        String failBPClassName = configuration.getAttribute(
                IDDLaunchConstants.FAILING_BP_CLASS_NAME, "");
        int failBPLineNumber = configuration.getAttribute(
                IDDLaunchConstants.FAILING_BP_LINE_NUMBER, -1);
        int failBPHitCount = configuration.getAttribute(
                IDDLaunchConstants.FAILING_BP_HIT_COUNT, 0);

        final boolean automaticMatchPointFinderEnabled = configuration
                .getAttribute(IDDLaunchConstants.ENABLE_MATCHPOINT_FINDER, true);

        // build the test description for the delta debugging run
        IJavaProject javaProject = this.getJavaProject(configuration);
        IVMInstall vmInstall = this.getVMInstall(configuration);

        String passTestClassName = this.getTestClassName(configuration);
        String failTestClassName = passTestClassName;

        IJUnitTestDescription passJUnitTestDesc = new JUnitTestDescription(
                passTestClassName, passTestName);
        IJUnitTestDescription failJUnitTestDesc = new JUnitTestDescription(
                failTestClassName, failTestName);

        IMatchPoint matchPoint = null;

        if (!automaticMatchPointFinderEnabled) {
            // add breakpoints
            IBreakpointLocation passBPLocation = new LineBreakpointLocation(
                    passBPClassName, passBPLineNumber, passBPHitCount);
            IBreakpointLocation failBPLocation = new LineBreakpointLocation(
                    failBPClassName, failBPLineNumber, failBPHitCount);

            matchPoint = new MatchPoint(passBPLocation, failBPLocation);

        }

        final IDDLaunchDescription ddLaunchDesc = new DDLaunchDescription(
                javaProject, vmInstall, passJUnitTestDesc, failJUnitTestDesc,
                matchPoint);

        // if automatic match point finder is enabled use the CTS launcher.
        // Otherwise we use the default launcher (this is SimpleDDLauncher).

        Job ddStateRunJob = new Job("DDstate") {

            protected IStatus run(IProgressMonitor pmonitor) {
                IDDLauncher launcher = null;
                if (!automaticMatchPointFinderEnabled) {
                    launcher = DDStateLaunchPlugin.getDDLauncher();
                    launcher.disableMatchPointFinder();
                } else {
                    launcher = new CTSLauncher();
                    launcher.enableMatchPointFinder();
                }

                try {
                    pmonitor.beginTask(
                            "Compute failure-inducing state changes", 100);
                    launcher.start(ddLaunchDesc, new SubProgressMonitor(
                            pmonitor, 100));
                } finally {
                    pmonitor.done();
                }

                return Status.OK_STATUS;
            }

        };

        ddStateRunJob.setUser(true);
        ddStateRunJob.setSystem(false);
        ddStateRunJob.setPriority(Job.LONG);
        ddStateRunJob.schedule();

    }

    /**
     * Returns whether a launch should proceed. This method is called first in
     * the launch sequence providing an opportunity for this launch delegate to
     * abort the launch.
     * 
     * In special this mehtod checks if the launch configuration contains the
     * line numbers for the passing and the failing breakpoint.
     * 
     * @param configuration
     *            configuration being lanuched
     * @param mode
     *            launch mode
     * @param monitor
     *            progress monitor
     * @return whether the launch should proceed
     * @throws CoreException
     *             if an exception occurs while performing pre launch checks
     */
    public boolean preLaunchCheck(final ILaunchConfiguration configuration,
            final String mode, IProgressMonitor monitor) throws CoreException {

        boolean result = true;

        String passTestName = configuration.getAttribute(
                IDDLaunchConstants.PASSING_TEST_NAME, "");
        String failTestName = configuration.getAttribute(
                IDDLaunchConstants.FAILING_TEST_NAME, "");

        result &= !failTestName.equals("");
        result &= !passTestName.equals("");
        result &= !failTestName.equals(passTestName);

        boolean automaticMatchPointFinderEnabled = configuration.getAttribute(
                IDDLaunchConstants.ENABLE_MATCHPOINT_FINDER, true);

        if (!automaticMatchPointFinderEnabled) {
            int passBPLineNumber = configuration.getAttribute(
                    IDDLaunchConstants.PASSING_BP_LINE_NUMBER, -1);
            int passBPHitCount = configuration.getAttribute(
                    IDDLaunchConstants.PASSING_BP_HIT_COUNT, 0);
            String passBPClassName = configuration.getAttribute(
                    IDDLaunchConstants.PASSING_BP_CLASS_NAME, "");

            int failBPLineNumber = configuration.getAttribute(
                    IDDLaunchConstants.FAILING_BP_LINE_NUMBER, -1);
            int failBPHitCount = configuration.getAttribute(
                    IDDLaunchConstants.FAILING_BP_HIT_COUNT, 0);
            String failBPClassName = configuration.getAttribute(
                    IDDLaunchConstants.FAILING_BP_CLASS_NAME, "");

            result &= passBPLineNumber > 0;
            result &= failBPLineNumber > 0;
            result &= passBPHitCount > 0;
            result &= failBPHitCount > 0;
            result &= !passBPClassName.equals("");
            result &= !failBPClassName.equals("");
        }

        if (result == false) {
            final IStatus warning = logger.log(IStatus.WARNING,
                    "The given launch configuration is not valid. "
                            + "Look at your launch configuration dialog "
                            + "for the error messages.");

            DDStateLaunchPlugin.getDisplay().asyncExec(new Runnable() {

                public void run() {
                    IWorkbenchWindow window = PlatformUI.getWorkbench()
                            .getActiveWorkbenchWindow();

                    ILaunchGroup group = DebugUITools.getLaunchGroup(
                            configuration, mode);
                    if (group != null) {
                        DebugUITools.openLaunchConfigurationDialog(window
                                .getShell(), configuration, group
                                .getIdentifier(), warning);
                    }
                }
            });
        }

        return result && super.preLaunchCheck(configuration, mode, monitor);
    }

    /**
     * Return the full qualified name of the test type defined in the given
     * launch configuration.
     * 
     * @param configuration
     *            launch configuration for the delta debugging run.
     * @return the full qualified name of the test type defined in the launch
     *         configuration.
     * @throws CoreException
     *             if this method fails. Reasons include:
     *             <ul>
     *             <li>An exception occured while retrieving the type names of
     *             the given launch configuration.</li>
     *             <li>Not exact one test type is defined in the given launch
     *             configuration.</li>
     *             </ul>
     */
    private String getTestClassName(ILaunchConfiguration configuration)
            throws CoreException {
        String result = null;

        result = configuration
                .getAttribute(
                        org.eclipse.jdt.launching.IJavaLaunchConfigurationConstants.ATTR_MAIN_TYPE_NAME,
                        (String) null);

        return result;
    }
}