/*
 * $Id: AbstractClassInstrumenter.java,v 1.1 2005/07/21 12:46:51 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Jul 18, 2005 10:47:09 AM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.bcel.instrumenter;

import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.InstructionFactory;

/**
 * COMMENTME Comment type AbstractClassInstrumenter
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class AbstractClassInstrumenter implements IClassInstrumenter {

    /**
     * Set debug to true to provide additional informations for the class to
     * instrument.
     */
    private boolean fDebug = false;

    /**
     * All packages and classes, which should not instrumet by the class
     * instrumenter.
     */
    private String[] fIgnoredPackages = null;

    public AbstractClassInstrumenter(String[] ignoredPackages) {
        this.fIgnoredPackages = ignoredPackages;
    }

    public AbstractClassInstrumenter() {
        this.fIgnoredPackages = new String[0];
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.bcel.instrumenter.IClassInstrumenter#modifyClass(org.apache.bcel.classfile.JavaClass)
     */
    public JavaClass modifyClass(JavaClass clazz)
            throws InstrumentationException {
        ConstantPoolGen constantPoolGenerator = new ConstantPoolGen(clazz
                .getConstantPool());
        ClassGen classGenerator = new ClassGen(clazz);
        classGenerator.setConstantPool(constantPoolGenerator);
        InstructionFactory factory = new InstructionFactory(classGenerator,
                constantPoolGenerator);

        if (!clazz.isInterface()
                && !this.checkIfClassIsInIgnoredPackages(clazz.getClassName())) {

            clazz = this.modifyClass(clazz, constantPoolGenerator,
                    classGenerator, factory);

            clazz.setConstantPool(constantPoolGenerator.getFinalConstantPool());
            clazz = classGenerator.getJavaClass();
        }

        return clazz;
    }

    public abstract JavaClass modifyClass(JavaClass clazz,
            ConstantPoolGen constantPoolGenerator, ClassGen classGenerator,
            InstructionFactory factory) throws InstrumentationException;

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.bcel.instrumenter.IClassInstrumenter#getIgnoredPackages()
     */
    public String[] getIgnoredPackages() {
        return this.fIgnoredPackages;
    }

    /**
     * Set the packages that should be ignored by the class instrumenter.
     * 
     * @param ignoredPackages
     *            the packages that should be ignored by the class instrumenter.
     */
    public void setIgnoredPackages(String[] ignoredPackages) {
        this.fIgnoredPackages = ignoredPackages;
    }

    /**
     * Add the given package name to the packages, that should be ignored by the
     * class instrumenter.
     * 
     * @param packageName
     *            the name of the package that should be ignored by the class
     *            instrumenter.
     */
    public void addPackageToIgnore(String packageName) {
        int iPackageLength = this.fIgnoredPackages.length;
        String[] result = new String[iPackageLength + 1];

        System.arraycopy(this.fIgnoredPackages, 0, result, 0, iPackageLength);
        result[iPackageLength] = packageName;

        this.fIgnoredPackages = result;
    }

    /**
     * Add the given packages names to the packages, that should be ignored by
     * the class instrumenter.
     * 
     * @param packageNames
     *            the name of the packages that should be ignored by the class
     *            instrumeter;
     */
    public void addPackagesToIgnore(String[] packageNames) {
        int iPackageLength = this.fIgnoredPackages.length;
        int newPackageLength = packageNames.length;

        String[] result = new String[iPackageLength + newPackageLength];
        System.arraycopy(this.fIgnoredPackages, 0, result, 0, iPackageLength);
        System.arraycopy(packageNames, 0, result, iPackageLength,
                newPackageLength);

        this.fIgnoredPackages = result;
    }

    /**
     * Return true if the debug mode is enabled, otherwise false.
     * 
     * @return true if the debug mode is enabled, otherwise false.
     */
    public boolean isDebug() {
        return this.fDebug;
    }

    /**
     * Check if the given class name match at least one of the entries of the
     * ignored packages.
     * 
     * <b>Note: </b> It satisfy if the given class starts with at least one
     * entry of the ignored packages.
     * 
     * @param className
     *            the name of the class that should be check if it is in the
     *            ignored packages.
     * @return true, if the given class name match at least one of the entries
     *         of the ignored packages, otherwise false.
     */
    public boolean checkIfClassIsInIgnoredPackages(String className) {
        boolean result = false;

        for (int i = 0; i < this.fIgnoredPackages.length; i++) {
            if (className.startsWith(this.fIgnoredPackages[i])) {
                result = true;
                break;
            }
        }

        return result;
    }

    /**
     * Use this message to print all methods of the given class generator.
     * 
     * @param classGenerator
     */
    protected void printAllMethods(ClassGen classGenerator) {
        if (this.fDebug && classGenerator != null) {
            System.out.println("Methods ---");
            Method[] methods = classGenerator.getMethods();
            for (int i = 0; i < methods.length; i++) {
                Method method = methods[i];
                System.out.println("  " + method);
            }
            System.out.println("--- Methods");
        }
    }

}
