/*
 * $Id: AbstractLauncher.java,v 1.5 2005/08/11 13:41:31 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 28.07.2005 13:06:49 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client;

import org.deltadebugging.ddstate.comparator.GraphComparator;
import org.deltadebugging.ddstate.comparator.IGraphComparator;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.matcher.IMatchPointFinder;

/**
 * The AbstractLauncher provide the base functionality for a delta debugging
 * laucnher. Every delta debugging launcher should extend the AbstractLauncher.
 * Here you can set the preferred GraphComparator, MatchPointFinder etc.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class AbstractLauncher implements IDDLauncher {

	/**
	 * This is a counter fo the DDLaunchID. We use the DDLaunchID to identify
	 * every delta debugging launch.
	 */
	private static int fNextDDLaunchID = 1;

	/**
	 * This is a counter for the DDRunID. We use the DDRunID to identify every
	 * delta debugging run.
	 */
	private static int fNextDDRunID = 1;

	/** The graph comparator used to compare the two memory graphs */
	private IGraphComparator fGraphComparator = null;

	/**
	 * The match point finder used to find all match points of the passing and
	 * the failing run
	 */
	private IMatchPointFinder fMatchPointFinder = null;

	/** Flag to enabled or disable the match point finder */
	private boolean fMatchPointFinderEnabled = true;

	/** The launch desc that describe what to launch */
	private IDDLaunchDescription fLaunchDesc = null;

	/** The test runner for the passing run */
	private IExtTestRunner fPassTestRunner = null;

	/** The test runner for the failing run */
	private IExtTestRunner fFailTestRunner = null;

	public AbstractLauncher(IMatchPointFinder matchPointFinder) {
		this.fMatchPointFinder = matchPointFinder;
		this.fGraphComparator = GraphComparator.getInstance();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#setLaunchDescription(org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription)
	 */
	public void setLaunchDescription(IDDLaunchDescription launchDesc) {
		this.fLaunchDesc = launchDesc;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#setFailTestRunner(org.deltadebugging.ddstate.launcher.client.IExtTestRunner)
	 */
	public void setFailTestRunner(IExtTestRunner failTestRunner) {
		this.fFailTestRunner = failTestRunner;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#setPassTestRunner(org.deltadebugging.ddstate.launcher.client.IExtTestRunner)
	 */
	public void setPassTestRunner(IExtTestRunner passTestRunner) {
		this.fPassTestRunner = passTestRunner;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#getLaunchDescription()
	 */
	public IDDLaunchDescription getLaunchDescription() {
		return this.fLaunchDesc;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#getFailTestRunner()
	 */
	public IExtTestRunner getFailTestRunner() {
		return this.fFailTestRunner;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#getPassTestRunner()
	 */
	public IExtTestRunner getPassTestRunner() {
		return this.fPassTestRunner;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#isMatchPointFinderEnabled()
	 */
	public boolean isMatchPointFinderEnabled() {
		return this.fMatchPointFinderEnabled;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#enableMatchPointFinder()
	 */
	public void enableMatchPointFinder() {
		this.fMatchPointFinderEnabled = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#disableMatchPointFinder()
	 */
	public void disableMatchPointFinder() {
		this.fMatchPointFinderEnabled = false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#setMatchPointFinder(org.deltadebugging.ddstate.launcher.matcher.IMatchPointFinder)
	 */
	public void setMatchPointFinder(IMatchPointFinder matchPointFinder) {
		if (matchPointFinder == null)
			throw new IllegalArgumentException(
					"Given match point finder must not be null");

		this.fMatchPointFinder = matchPointFinder;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#setGraphComparator(org.deltadebugging.ddstate.comparator.IGraphComparator)
	 */
	public void setGraphComparator(IGraphComparator graphComparator) {
		if (graphComparator == null)
			throw new IllegalArgumentException(
					"Given graph comparator must not be null");

		this.fGraphComparator = graphComparator;
	}

	/**
	 * Generate a new id for a delta debugging run.
	 * 
	 * @return unique id for a delta debugging run.
	 */
	protected synchronized int getNextDDRunID() {
		return fNextDDRunID++;
	}

	/**
	 * Generate a new id for a delta debugging launch.
	 * 
	 * @return unique id for a delta debugging launch.
	 */
	protected synchronized int getNextDDLaunchID() {
		return fNextDDLaunchID++;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#getGraphComparator()
	 */
	public IGraphComparator getGraphComparator() {
		return this.fGraphComparator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#getMatchPointFinder()
	 */
	public IMatchPointFinder getMatchPointFinder() {
		return this.fMatchPointFinder;
	}
}
