/*
 * $Id: JUnitTestLauncher.java,v 1.10 2005/11/17 12:32:18 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 08.01.2005 12:08:51 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.client.testrunner.TestRunnerDelegate;
import org.deltadebugging.ddstate.launcher.server.IRemoteTestManager;
import org.deltadebugging.ddstate.launcher.server.IRemoteTestRunner;
import org.deltadebugging.ddstate.launcher.server.RemoteTestManager;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationType;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.core.model.IDebugTarget;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.internal.junit.ui.JUnitPlugin;
import org.eclipse.jdt.launching.IJavaLaunchConfigurationConstants;
import org.eclipse.jdt.launching.IRuntimeClasspathEntry;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.jdt.launching.JavaRuntime;
import org.eclipse.jdt.launching.SocketUtil;

/**
 * COMMENTME Comment type JUnitTestLauncher
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class JUnitTestLauncher {

	// Personal logger for this class
	final static DDLogger logger = DDStateLaunchPlugin
			.getLogger(JUnitTestLauncher.class);

	private final static String ATTR_LAUNCH_ID = DDStateLaunchPlugin.PLUGIN_ID
			+ ".ATTR_LAUNCH_ID";

	boolean fStarted = false;

	private int fPort = -1;

	IRemoteTestManager fTestManager = null;

	private int fLaunchID = -1;

	private IJavaDebugTarget fDebugTarget = null;

	private static int fLaunchIDCounter = 0;

	private final static long MAX_WAIT_TIME_CONNECT = 60000; // ms

	private final static long TIME_INTERVALL_CONNECT = 200; // ms

	private synchronized int getNextLaunchID() {
		return JUnitTestLauncher.fLaunchIDCounter++;
	}

	public synchronized void start(IJavaProject project, IVMInstall vmInstall,
			final IProgressMonitor monitor) throws CoreException {
		if (this.fStarted)
			return;

		// get the next launch id
		this.fLaunchID = this.getNextLaunchID();

		final ILaunchConfigurationWorkingCopy configuration = this
				.createLaunchConfiguration();

		// specify the installed jre
		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_VM_INSTALL_NAME,
				vmInstall.getName());
		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_VM_INSTALL_TYPE,
				vmInstall.getVMInstallType().getId());

		// set the corresponding project
		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_PROJECT_NAME, project
						.getElementName());

		// set the main type
		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_MAIN_TYPE_NAME,
				RemoteTestManager.class.getName());

		// set the classpath
		List classPath = this.createBootstrapClassPath(vmInstall);
		classPath.addAll(this.createClassPath(configuration));

		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_CLASSPATH, classPath);
		configuration
				.setAttribute(
						IJavaLaunchConfigurationConstants.ATTR_DEFAULT_CLASSPATH,
						false);

		// set the fPort
		this.fPort = SocketUtil.findFreePort();
		configuration.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_PROGRAM_ARGUMENTS,
				"-Port " + this.fPort);

		// set the unique launch id - this is necessary to find the associated
		// debug target of the launch
		configuration.setAttribute(ATTR_LAUNCH_ID, this.fLaunchID);

		DDStateLaunchPlugin.getDisplay().syncExec(new Runnable() {
			public void run() {
                try {
                    configuration.launch(ILaunchManager.DEBUG_MODE, new NullProgressMonitor());
                } catch (CoreException e) {
                    JUnitTestLauncher.logger.log(IStatus.ERROR, "Error launching junit test", e);
                    monitor.setCanceled(true);
                }
			}
		});

		new MonitorObserverJob("Observe canceld monitor", this, monitor)
				.schedule();

		// connect to the server and get the test manager
		this.fTestManager = this.connectToServer("localhost", this.fPort,
				monitor);

		// find debug target for the launch id
		if (!monitor.isCanceled()) {
			this.fDebugTarget = findDebugTargetForID(this.fLaunchID);

			assert this.fDebugTarget != null;
		}

		if (monitor.isCanceled()) {
			this.shutDown(false);
			return;
		}

		this.fStarted = true;
	}

	/**
	 * This job observe the given monitor. If the monitor was canceld and the
	 * test manager of the given test launcher is always running. The job
	 * rescheduled itself after 3 seconds. If the test launcher was stopped the
	 * job exit.
	 * 
	 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
	 */
	class MonitorObserverJob extends Job {

		/**
		 * The monitor to observe.
		 */
		private final IProgressMonitor fMonitor;

		private final JUnitTestLauncher fTestLauncher;

		public MonitorObserverJob(String name, JUnitTestLauncher testLauncher,
				IProgressMonitor monitor) {
			super(name);
			this.setSystem(true);
			this.setPriority(Job.LONG);
			this.fMonitor = monitor;
			this.fTestLauncher = testLauncher;
		}

		protected IStatus run(IProgressMonitor monitor) {

			if (this.fMonitor != null && this.fMonitor.isCanceled()
					&& this.fTestLauncher.isStarted()) {
				new ShutdownJob("Force shutdown test manager",
						this.fTestLauncher, false).schedule();
			}

			// schedule after 1 second
			if (monitor != null && this.fTestLauncher.isStarted()) {
				this.schedule(3000);
			}

			return Status.OK_STATUS;
		}

	}

	private static IJavaDebugTarget findDebugTargetForID(int launchID)
			throws CoreException {
		IJavaDebugTarget result = null;

		ILaunchManager launchMngr = DebugPlugin.getDefault().getLaunchManager();
		IDebugTarget[] debugTargets = launchMngr.getDebugTargets();

		for (int i = 0; i < debugTargets.length; i++) {
			IDebugTarget target = debugTargets[i];

			ILaunch launch = target.getLaunch();
			if (launch != null) {
				ILaunchConfiguration launchConfig = launch
						.getLaunchConfiguration();
				int targetLaunchID = launchConfig.getAttribute(ATTR_LAUNCH_ID,
						-999);

				if (targetLaunchID == launchID) {
					assert target instanceof IJavaDebugTarget;

					result = (IJavaDebugTarget) target;
					break;
				}
			}

		}

		return result;
	}

	/**
	 * This job shutdown the test manager if it is already running. If it cannot
	 * shutdown the test manager within a given time the job forces the
	 * shutdown.
	 * 
	 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
	 */
	class ShutdownJob extends Job {

		private final static long MAX_WAIT_TIME_SHUTDOWN = 60000; // ms

		private final static long TIME_INTERVALL_SHUTDOWN = 50; // ms

		private final JUnitTestLauncher fTestLauncher;

		private final boolean fForceShutDown;

		public ShutdownJob(String name, JUnitTestLauncher testLauncher,
				boolean forceShutDown) {
			super(name);
			this.fTestLauncher = testLauncher;
			this.fForceShutDown = forceShutDown;
		}

		protected IStatus run(IProgressMonitor monitor) {

			if (this.fTestLauncher.isStarted()) {
				this.fTestLauncher.fStarted = false;

				if (!this.fForceShutDown) {
					long maxNumberAttemps = MAX_WAIT_TIME_SHUTDOWN
							/ TIME_INTERVALL_SHUTDOWN;

					long i = 0;

					try {
						while (this.fTestLauncher.fTestManager
								.hasRunningTests()
								&& i < maxNumberAttemps) {
							try {
								Thread.sleep(TIME_INTERVALL_SHUTDOWN);

							} catch (InterruptedException e) {
								// do nothing

							} finally {
								i++;
							}
						}
					} catch (RemoteException e) {
						// fall through
					}
				}

				try {
					this.fTestLauncher.fTestManager.shutDown();
				} catch (RemoteException e) {
					// fall through
				}

				// reset variables
				this.fTestLauncher.fTestManager = null;
			}

			return Status.OK_STATUS;
		}

	}

	public synchronized boolean isStarted() {
		return this.fStarted;
	}

	public synchronized void shutDown(boolean forceShutDown) {

		if (this.isStarted()) {

			if (this.fTestManager != null) {
				new ShutdownJob("Shutdown test manager", this, forceShutDown)
						.schedule();
			}

			// reset the attributes of this test launcher
			this.fDebugTarget = null;
			this.fLaunchID = -1;
			this.fPort = -1;
		}
	}

	/**
	 * Try to connect the the rmi server on the given host on the given fPort.
	 * Return the remote test manager, which have to be available in the jndi
	 * context.
	 * 
	 * @param rmiHost
	 * @param rmiPort
	 * @param monitor
	 *            the progress monitor.
	 * @return remote test manager.
	 * @throws CoreException
	 *             if it is not possible to establish a connection to the rmi
	 *             server.
	 */
	private IRemoteTestManager connectToServer(String rmiHost, int rmiPort,
			IProgressMonitor monitor) throws CoreException {
		IRemoteTestManager result = null;

		long maxNumberAttemps = MAX_WAIT_TIME_CONNECT / TIME_INTERVALL_CONNECT;

		String url = RemoteTestManager.createURL(rmiHost, rmiPort);
		String errorMsg = "Cannot connect to DDRemoteTestManager with URL '"
				+ url + "'";

		for (long i = 0; i < maxNumberAttemps && result == null; i++) {

			try {
				if (monitor.isCanceled())
					return result;

				result = (IRemoteTestManager) Naming.lookup(url);

				logger.trace(DDStateLaunchPlugin.DEBUG,
						"Try to connect to DDRemoteTestManager ... OK");

			} catch (Exception e) {
				if (i == maxNumberAttemps) {
					IStatus status = logger.log(IStatus.ERROR, errorMsg, e);
					throw new CoreException(status);
				}

				try {
					Thread.sleep(TIME_INTERVALL_CONNECT);
				} catch (InterruptedException ie) {
					// do nothing
				}
			}
		}

		if (result == null) {
			IStatus status = logger.log(IStatus.ERROR, errorMsg);
			throw new CoreException(status);
		}

		return result;
	}

	public synchronized IExtTestRunner getTestRunner() throws CoreException {
		IExtTestRunner result = null;
		IRemoteTestRunner remoteTestRunner = null;

		boolean debugMode = DDStateLaunchPlugin.DEBUG_REMOTE_TEST_RUNNER;

		if (this.fTestManager != null) {
			try {
				remoteTestRunner = this.fTestManager.getTestRunner(debugMode);

				result = new TestRunnerDelegate(remoteTestRunner,
						this.fDebugTarget);
			} catch (Throwable t) {
				IStatus status = logger.log(IStatus.ERROR,
						"Cannot connect to the remote test runner", t);
				throw new CoreException(status);
			}
		}

		return result;
	}

	public synchronized IJavaDebugTarget getDebugTarget() {
		return this.fDebugTarget;
	}

	private ILaunchConfigurationWorkingCopy createLaunchConfiguration()
			throws CoreException {
		ILaunchConfigurationWorkingCopy result = null;

		ILaunchManager manager = DebugPlugin.getDefault().getLaunchManager();
		ILaunchConfigurationType type = manager
				.getLaunchConfigurationType(IJavaLaunchConfigurationConstants.ID_JAVA_APPLICATION);

		result = type.newInstance(null, "DDRemoteTestServer");

		return result;
	}

	private List createBootstrapClassPath(IVMInstall vmInstall)
			throws CoreException {
		List result = new ArrayList();

		URL ddstateLauncherURL = Platform.getBundle(
				DDStateLaunchPlugin.PLUGIN_ID).getEntry("/");
		URL bcelURL = Platform.getBundle("org.apache.bcel").getEntry("/");
		IPath modRTJarPath = RuntimeJarUtil.getFullModRTJarLocation(vmInstall);

		IRuntimeClasspathEntry entry = null;
		try {
			entry = JavaRuntime.newArchiveRuntimeClasspathEntry(modRTJarPath);
			entry
					.setClasspathProperty(IRuntimeClasspathEntry.BOOTSTRAP_CLASSES);
			result.add(entry.getMemento());

			entry = this.getClassPathEntry(ddstateLauncherURL,
					"ddstate_rt.jar", true);
			entry
					.setClasspathProperty(IRuntimeClasspathEntry.BOOTSTRAP_CLASSES);
			result.add(entry.getMemento());

			entry = this.getClassPathEntry(bcelURL, "bcel-5.1.jar");
			entry
					.setClasspathProperty(IRuntimeClasspathEntry.BOOTSTRAP_CLASSES);
			result.add(entry.getMemento());

		} catch (IOException e) {
			IStatus status = logger.log(IStatus.ERROR,
					"Cannot define the path for JUnit", e);
			throw new CoreException(status);
		}

		return result;
	}

	/**
	 * Create the classpath for the given configuration.
	 * 
	 * @param configuration
	 *            the launch configuration for the junit run.
	 * @return the classpath for the given configuration.
	 * @throws CoreException
	 */
	private List createClassPath(ILaunchConfiguration configuration)
			throws CoreException {

		IRuntimeClasspathEntry[] entries = JavaRuntime
				.computeUnresolvedRuntimeClasspath(configuration);
		entries = JavaRuntime.resolveRuntimeClasspath(entries, configuration);

		// add the user entries to the classpath
		List classPath = new ArrayList(entries.length);
		for (int i = 0; i < entries.length; i++) {
			if (entries[i].getClasspathProperty() == IRuntimeClasspathEntry.USER_CLASSES) {
				String location = entries[i].getMemento();
				if (location != null) {
					classPath.add(location);
				}
			}
		}

		URL url = Platform.getBundle(JUnitPlugin.PLUGIN_ID).getEntry("/"); //$NON-NLS-1$
		URL runtimeURL = Platform
				.getBundle("org.eclipse.jdt.junit.runtime").getEntry("/"); //$NON-NLS-1$ //$NON-NLS-2$
		URL ddstateLauncherURL = Platform.getBundle(
				DDStateLaunchPlugin.PLUGIN_ID).getEntry("/");
		URL ddstateCoreURL = Platform.getBundle(DDStatePlugin.PLUGIN_ID)
				.getEntry("/");

		try {
			classPath.add(this.getClassPathEntry(url, "junitsupport.jar")
					.getMemento());

			classPath.add(this
					.getClassPathEntry(runtimeURL, "junitruntime.jar")
					.getMemento());

			classPath.add(this.getClassPathEntry(ddstateLauncherURL,
					"ddstate_launcher.jar").getMemento());

			classPath.add(this.getClassPathEntry(ddstateCoreURL, "ddstate.jar")
					.getMemento());

		} catch (IOException e) {
			IStatus status = logger.log(IStatus.ERROR,
					"Cannot define the path for JUnit", e);
			throw new CoreException(status);
		}

		return classPath;
	}

	private IRuntimeClasspathEntry getClassPathEntry(URL url, String jarName)
			throws MalformedURLException, IOException {
		return this.getClassPathEntry(url, jarName, false);
	}

	private IRuntimeClasspathEntry getClassPathEntry(URL url, String jarName,
			boolean forcesJar) throws MalformedURLException, IOException {
		IRuntimeClasspathEntry result = null;

		IPath path = this.getPath(url, jarName, forcesJar);
		result = JavaRuntime.newArchiveRuntimeClasspathEntry(path);

		return result;
	}

	/**
	 * Return the path for the given url context and the given jar name.
	 * 
	 * @param url
	 * @param jarName
	 * @param forcesJar
	 *            forces the loading of the jar file not the development classes
	 * @return the path for the given url context and the given jar name.
	 * @throws MalformedURLException
	 *             if no protocol is specified, or an unknown protocol is found.
	 * @throws IOException
	 *             if unable to resolve URL
	 */
	private IPath getPath(URL url, String jarName, boolean forcesJar)
			throws MalformedURLException, IOException {
		IPath result = null;

		if (Platform.inDevelopmentMode() && !forcesJar) {

			try {
				result = new Path(Platform
						.asLocalURL(new URL(url, "bin")).getPath()); //$NON-NLS-1$
			} catch (IOException e3) {
				result = new Path(Platform.asLocalURL(new URL(url, jarName))
						.getPath()); //$NON-NLS-1$
			}
		} else {
			result = new Path(Platform.asLocalURL(new URL(url, jarName))
					.getPath());
		}

		return result;
	}
}