/*
 * $Id: CTSLauncher.java,v 1.24 2005/10/25 11:10:15 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 28.07.2005 16:33:48 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client.cts;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.comparator.exceptions.GraphComparatorException;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.client.BaseLauncher;
import org.deltadebugging.ddstate.launcher.client.LaunchExceptionHandler;
import org.deltadebugging.ddstate.launcher.client.TestRunnerException;
import org.deltadebugging.ddstate.launcher.client.exception.DeltaComputationException;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult;
import org.deltadebugging.ddstate.launcher.matcher.AllMethodCallMatcher;
import org.deltadebugging.ddstate.launcher.matcher.IMatchPoint;
import org.deltadebugging.ddstate.launcher.matcher.exception.MatchPointComputationException;
import org.deltadebugging.ddstate.launcher.tester.exception.TesterException;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;

/**
 * COMMENTME Comment type CTSLauncher
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class CTSLauncher extends BaseLauncher {

    private final static String INFO_MSG_NO_MATCHPOINT = "No suitable match points found. "
            + "Delta debugging cannot applied, because it exists no comparable points "
            + "between the defined test runs.";

    /** Personal logger for this class */
    private final static DDLogger logger = DDStateLaunchPlugin
            .getLogger(CTSLauncher.class);

    /**
     * Define the maximum search range for the search of a salvageable dd run
     * result.
     */
    private final static int MAX_SEARCH_RANGE = 3;

    public CTSLauncher() {
        super(new AllMethodCallMatcher());
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.BaseLauncher#startLauncherStrategy(org.eclipse.core.runtime.IProgressMonitor)
     */
    public void startLauncherStrategy(IProgressMonitor monitor) {

        try {
            monitor.beginTask("Start CTSLauncher", 100);

            //
            // notify all listeners the the cts run was started
            //
            DDStateLaunchPlugin.getCTSRunListenerManager().notifyCTSRunStarted(
                    this.getLaunchDescription());

            //
            // if the match point finder is set and enabled, start it and find
            // the comparable breakpoint in the test runs
            //
            IMatchPoint[] matchPoints = null;
            if (!monitor.isCanceled()) {
                if (this.getMatchPointFinder() != null
                        && this.isMatchPointFinderEnabled()) {
                    try {
                        matchPoints = this.getMatchPointFinder()
                                .computeMatchPoints(
                                        this.getLaunchDescription(),
                                        new SubProgressMonitor(monitor, 10));

                    } catch (MatchPointComputationException e) {
                        LaunchExceptionHandler.handleException(e);
                        monitor.setCanceled(true);
                    }
                }
            }

            // inform the user that delta debugging cannot applied
            if (matchPoints == null || matchPoints.length < 2) {
                LaunchExceptionHandler
                        .handleInformation(INFO_MSG_NO_MATCHPOINT);
                monitor.setCanceled(true);
            }

            if (!monitor.isCanceled()) {

                MatchPointInterval mpInterval = new MatchPointInterval(
                        matchPoints);

                CTSRunResult ctsResult = this.computeCTSResult(mpInterval,
                        new SubProgressMonitor(monitor, 90));

                if (ctsResult != null) {
                    // notify listener

                    for (int i = 0; i < ctsResult.getAllCauseTransitions().length; i++) {
                        IDDRunResult ddRunResult = ctsResult
                                .getAllCauseTransitions()[i];

                        DDStateLaunchPlugin.getDDRunListenerManager()
                                .notifyDDRunTerminated(ddRunResult);
                    }

                    //
                    // inform the listeners that the cts run has finished
                    //
                    DDStateLaunchPlugin.getCTSRunListenerManager()
                            .notifyCTSRunTerminated(ctsResult);
                }
            }

        } finally {
            monitor.done();
        }
    }

    /**
     * Compute the dd run result using the cts algorithm to find cause
     * transitions.
     * 
     * @param mpInterval
     * @param monitor
     * @return a list of cts
     */
    private CTSRunResult computeCTSResult(MatchPointInterval mpInterval,
            IProgressMonitor monitor) {

        CTSRunResult result = new CTSRunResult(this.getLaunchDescription(),
                mpInterval);
        SearchStrategy searchStrategy = new SearchStrategy(this,
                MAX_SEARCH_RANGE);

        try {
            this.computeDDRunResults(result, searchStrategy, mpInterval, true,
                    monitor);

        } catch (Throwable t) {
            LaunchExceptionHandler.handleException(t);

        }

        return result;
    }

    private void computeDDRunResults(CTSRunResult ctsResult,
            SearchStrategy searchStrategy, MatchPointInterval mpInterval,
            boolean firstCall, IProgressMonitor monitor)
            throws GraphComparatorException, DeltaComputationException,
            TestRunnerException, TesterException {

        SearchStrategy.SearchResult leftSearchResult = null;
        SearchStrategy.SearchResult rightSearchResult = null;
        IDDRunResult leftDDRunResult = null;
        IDDRunResult rightDDRunResult = null;

        try {
            monitor.beginTask("DDstate - search in time", 100);
            
            if (!monitor.isCanceled()) {
                int acceptedOutcomes = SearchStrategy.RELEVANT_DELTAS_FOUND;

                if (firstCall) {
                    acceptedOutcomes = acceptedOutcomes
                            | SearchStrategy.NO_RELEVANT_DELTAS_FOUND;
                }

                leftSearchResult = searchStrategy.searchSuccessfulDDRun(
                        mpInterval, true, acceptedOutcomes, new SubProgressMonitor(monitor, 25));

                if (leftSearchResult.searchOutcome != SearchStrategy.NO_OUTCOME) {
                    IMatchPoint searchPoint = mpInterval
                            .getMatchPointAtIndex(leftSearchResult.resultIndex);
                    leftDDRunResult = leftSearchResult.ddRunResult;
                    ctsResult.addDDRunResult(searchPoint, leftDDRunResult);
                }
            }

            if (leftSearchResult != null
                    && leftSearchResult.resultIndex + 1 < mpInterval.length()) {
                MatchPointInterval rightMpInterval = mpInterval.getSubInterval(
                        leftSearchResult.resultIndex + 1,
                        mpInterval.length() - 1);

                if (!monitor.isCanceled()) {
                    rightSearchResult = searchStrategy.searchSuccessfulDDRun(
                            rightMpInterval, false,
                            SearchStrategy.RELEVANT_DELTAS_FOUND, new SubProgressMonitor(monitor, 25));

                    if (rightSearchResult.searchOutcome != SearchStrategy.NO_OUTCOME) {
                        IMatchPoint searchPoint = rightMpInterval
                                .getMatchPointAtIndex(rightSearchResult.resultIndex);
                        rightDDRunResult = rightSearchResult.ddRunResult;
                        ctsResult.addDDRunResult(searchPoint, rightDDRunResult);
                    }
                }
            }

            boolean isCauseTransion = false;
            if (!monitor.isCanceled() && leftDDRunResult != null
                    && rightDDRunResult != null) {
                isCauseTransion = CTSLauncher.isCauseTransition(
                        leftDDRunResult, rightDDRunResult);
            }

            // if their exists a cause transition between the two computed
            // results
            // continue the search for father cause transitions
            if (isCauseTransion) {
                logger.trace(DDStateLaunchPlugin.DEBUG_CTS_LAUNCHER,
                        "Cause transition found");

                if (mpInterval.canDivideInterval()) {
                    MatchPointInterval[] mpIntervals = mpInterval
                            .divideInTwoIntervals();

                    this.computeDDRunResults(ctsResult, searchStrategy,
                            mpIntervals[0], false, new SubProgressMonitor(monitor, 25));
                    this.computeDDRunResults(ctsResult, searchStrategy,
                            mpIntervals[1], false, new SubProgressMonitor(monitor, 25));

                }
            } else {
                monitor.worked(50);
                
            }
            
        } finally {
            monitor.done();
        }
    }

    /**
     * Check if between the given two test results exists a cause transition.
     * 
     * @param firstTestResult
     *            the first test result.
     * @param secondTestResult
     *            the second test result.
     * @return true if between the given two test results exists a cause
     *         transition, otherwise false.
     */
    public static boolean isCauseTransition(IDDRunResult firstTestResult,
            IDDRunResult secondTestResult) {
        boolean result = true;

        DeltaSet passDeltaSet = firstTestResult.getMinimizedDeltaSet();
        DeltaSet failDeltaSet = secondTestResult.getMinimizedDeltaSet();

        IMemoryGraphDelta[] passDeltas = (IMemoryGraphDelta[]) passDeltaSet
                .toArray(new IMemoryGraphDelta[passDeltaSet.size()]);
        IMemoryGraphDelta[] failDeltas = (IMemoryGraphDelta[]) failDeltaSet
                .toArray(new IMemoryGraphDelta[failDeltaSet.size()]);

        if (passDeltas.length == failDeltas.length) {

            boolean isNotCauseTransition = true;

            for (int i = 0; i < passDeltas.length
                    && isNotCauseTransition == true; i++) {
                IMemoryGraphDelta passDelta = passDeltas[i];
                boolean equalDeltaFound = false;

                for (int j = 0; j < failDeltas.length
                        && equalDeltaFound == false; j++) {
                    IMemoryGraphDelta failDelta = failDeltas[j];

                    if (failDelta != null) {
                        if (passDelta.ctsEqual(failDelta)) {
                            equalDeltaFound = true;
                            failDeltas[j] = null;
                        }

                    }
                }

                isNotCauseTransition &= equalDeltaFound;
            }

            result = !isNotCauseTransition;
        }

        return result;
    }

}
