/*
 * $Id: TestRunnerDelegate.java,v 1.14 2005/11/28 15:28:44 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 02.02.2005 08:42:38 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client.testrunner;

import java.rmi.RemoteException;
import java.rmi.server.RemoteObject;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.deltadebugging.ddstate.extractor.MemoryAccessUtil;
import org.deltadebugging.ddstate.extractor.exceptions.MemoryAccessException;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.client.IExtTestRunner;
import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;
import org.deltadebugging.ddstate.launcher.client.ObjectCreationException;
import org.deltadebugging.ddstate.launcher.client.TestRunnerException;
import org.deltadebugging.ddstate.launcher.matcher.MethodTrace;
import org.deltadebugging.ddstate.launcher.server.GenerateTraceException;
import org.deltadebugging.ddstate.launcher.server.IRemoteTestRunner;
import org.deltadebugging.ddstate.launcher.server.JUnitTestRunner;
import org.deltadebugging.ddstate.launcher.server.instrumenter.IdentifierMap;
import org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.GraphElementUtil;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.RootNode;
import org.deltadebugging.ddstate.memorygraph.exceptions.MemoryGraphExtractionException;
import org.deltadebugging.ddstate.memorygraph.type.IReferenceTypeNode;
import org.deltadebugging.ddstate.memorygraph.type.ITypeNode;
import org.deltadebugging.ddstate.memorygraph.util.Path;
import org.deltadebugging.ddstate.memorygraph.value.IArrayNode;
import org.deltadebugging.ddstate.memorygraph.value.IClassObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IStringObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IValueNode;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaFieldVariable;
import org.eclipse.jdt.debug.core.IJavaReferenceType;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaValue;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * COMMENTME Comment type DDRemoteTestRunnerDelegate
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class TestRunnerDelegate extends RemoteObject implements IExtTestRunner {

    private static final long serialVersionUID = 1353657581429683834L;

    /** Personal logger for this class */
    final static DDLogger logger = DDStateLaunchPlugin
            .getLogger(TestRunnerDelegate.class);

    private final IRemoteTestRunner fRemoteTestRunner;

    private final Map fCreatedObjects;

    private final IJavaDebugTarget fDebugTarget;

    public TestRunnerDelegate(IRemoteTestRunner remoteTestRunner,
            IJavaDebugTarget debugTarget) {
        assert debugTarget != null;

        this.fRemoteTestRunner = remoteTestRunner;

        this.fCreatedObjects = new HashMap();
        this.fDebugTarget = debugTarget;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#getDebugTarget()
     */
    public IJavaDebugTarget getDebugTarget() {
        return this.fDebugTarget;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#run(org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription)
     */
    public IPredicate run(IJUnitTestDescription testDesc)
            throws TestRunnerException {
        if (testDesc == null)
            throw new IllegalArgumentException(
                    "Given test description must not be null");

        // clear the maps of the created objects and the created arrays
        this.fCreatedObjects.clear();

        try {
            return this.fRemoteTestRunner.run(testDesc);
        } catch (RemoteException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on running test.", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.IExtTestRunner#generateTrace(org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription,
     *      org.deltadebugging.ddstate.launcher.server.instrumenter.IdentifierMap)
     */
    public MethodTrace generateTrace(IJUnitTestDescription testDesc,
            IdentifierMap identifierMap) throws GenerateTraceException,
            TestRunnerException {
        if (testDesc == null)
            throw new IllegalArgumentException(
                    "Given test description must not be null");

        try {
            return this.fRemoteTestRunner
                    .generateTrace(testDesc, identifierMap);
        } catch (RemoteException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on running test.", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#getCurrentClassLoaderID()
     */
    public int getCurrentClassLoaderID() throws TestRunnerException {
        try {
            return this.fRemoteTestRunner.getCurrentClassLoaderID();
        } catch (RemoteException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on running test.", e);
        }
    }

    private Path findNewCreatedObject(String className, int index,
            IJavaThread thread) throws TestRunnerException {
        Path result = null;

        if (className == null)
            throw new IllegalArgumentException(
                    "Given class name must not be null");

        if (thread == null)
            throw new IllegalArgumentException("Given thread must not be null");

        try {
            // We have to look for
            // org.deltadebugging.ddstate.launcher.server.JUnitTestRunner.newObjects[index]
            IJavaReferenceType refType = MemoryAccessUtil
                    .searchReferenceTypeByName(JUnitTestRunner.class.getName(),
                            this.getCurrentClassLoaderID(), thread);

            RootNode rootNode = new RootNode(true, -1);
            ITypeNode testRunnerNode = GraphElementUtil.createTypeNode(
                    rootNode, refType);
            IGraphEdge testRunnerEdge = GraphElementUtil.createTypeEdge(
                    rootNode, (IReferenceTypeNode) testRunnerNode);

            IJavaFieldVariable field = refType.getField("newObjects");
            IArrayNode newObjectsNode = (IArrayNode) GraphElementUtil
                    .createValueNode(rootNode, field);
            IGraphEdge newObjectsEdge = GraphElementUtil
                    .createGlobalVariableEdge(
                            (IReferenceTypeNode) testRunnerNode,
                            newObjectsNode, "newObjects");

            IJavaValue newObjects = (IJavaValue) field.getValue().getAdapter(
                    IJavaValue.class);
            IJavaVariable newJavaVariable = (IJavaVariable) newObjects
                    .getVariables()[index].getAdapter(IJavaVariable.class);
            IValueNode newVariableNode = GraphElementUtil.createValueNode(
                    rootNode, newJavaVariable);
            IGraphEdge newVariableEdge = GraphElementUtil
                    .createArrayElementEdge(newObjectsNode, newVariableNode,
                            index);

            // create Path - use this workaround with the list to avoid the
            // security checks at the path
            List edgeList = new LinkedList();
            edgeList.add(testRunnerEdge);
            edgeList.add(newObjectsEdge);
            edgeList.add(newVariableEdge);

            result = new Path(edgeList);

        } catch (MemoryAccessException e) {
            throw new TestRunnerException(
                    "Cannot retrieve the new created value.", e);
        } catch (DebugException e) {
            throw new TestRunnerException(
                    "Cannot retrieve the new created value.", e);
        } catch (MemoryGraphExtractionException e) {
            throw new TestRunnerException(
                    "Cannot retrieve the new created value.", e);
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#createNewObject(org.deltadebugging.ddstate.memorygraph.value.IObjectNode,
     *      org.eclipse.jdt.debug.core.IJavaThread)
     */
    public IJavaValue createNewObject(IObjectNode objToCreate,
            IJavaThread thread) throws ObjectCreationException,
            TestRunnerException {
        IJavaValue result = null;

        if (objToCreate == null)
            throw new IllegalArgumentException(
                    "Given object to create must not be null");

        if (thread == null)
            throw new IllegalArgumentException("Given thread must not be null");

        try {
            String className = objToCreate.getReferenceTypeName();
            int index = -1;

            // we have to distinguish between arrays and other objects
            if (objToCreate instanceof IArrayNode) {
                int length = ((IArrayNode) objToCreate).getLength();
                index = this.fRemoteTestRunner
                        .createNewArray(className, length);

            } else if (objToCreate instanceof IStringObjectNode) {
                String string = ((IStringObjectNode) objToCreate)
                        .getStringValue();
                index = this.fRemoteTestRunner.createNewString(string);

            } else if (objToCreate instanceof IClassObjectNode) {
                String instanceType = ((IClassObjectNode) objToCreate)
                        .getInstanceType();
                index = this.fRemoteTestRunner.createNewClass(instanceType);

            } else {
                index = this.fRemoteTestRunner.createNewObject(className);

            }

            Path pathToObject = this.findNewCreatedObject(className, index,
                    thread);

            this.fCreatedObjects.put(objToCreate, pathToObject);

            result = (IJavaValue) MemoryAccessUtil.getVariable(thread,
                    pathToObject, this.getCurrentClassLoaderID()).getValue()
                    .getAdapter(IJavaValue.class);

        } catch (RemoteException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on creating new object.", e);
        } catch (DebugException e) {
            throw new ObjectCreationException(
                    "Cannot retrieve value of the created object", e);
        } catch (MemoryAccessException e) {
            throw new ObjectCreationException(
                    "Cannot find variable to the created object.", e);
        } catch (TestRunnerException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on creating new object.", e);
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#isObjectAlreadyCreated(org.deltadebugging.ddstate.memorygraph.value.IObjectNode)
     */
    public boolean isObjectAlreadyCreated(IObjectNode obj) {
        boolean result = false;

        if (this.fCreatedObjects != null) {
            result = this.fCreatedObjects.containsKey(obj);
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#getPathToCreatedObject(org.deltadebugging.ddstate.memorygraph.value.IObjectNode)
     */
    public Path getPathToCreatedObject(IObjectNode obj) {
        Path result = null;

        if (this.fCreatedObjects != null) {
            // return a copy - if someone make changes on the returned graph
            // this should nothing change at the stored graph
            result = new Path((Path) this.fCreatedObjects.get(obj));
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.ITestRunner#loadClass(org.deltadebugging.ddstate.memorygraph.type.IReferenceTypeNode)
     */
    public void loadClass(IReferenceTypeNode classToLoad)
            throws TestRunnerException, ClassNotFoundException {

        if (classToLoad == null)
            throw new IllegalArgumentException(
                    "Given class to load must not be null");

        try {
            this.fRemoteTestRunner.loadClass(classToLoad.getTypeName());
        } catch (RemoteException e) {
            throw new TestRunnerException(
                    "Unexpected Exception on creating new object.", e);
        }
    }
}