/*
 * $Id: IDDRunResult.java,v 1.10 2005/08/19 15:18:34 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 10.04.2005 18:20:37 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.ddrun;

import java.util.Collection;
import java.util.Date;

import org.deltadebugging.ddcore.DeltaSet;
import org.eclipse.core.runtime.IStatus;

/**
 * The DDRunResult is the result of one delta debugging run. But it does not
 * only contains the result by itself. The result holds informations about start
 * and stop time, computed delta set, passing and failing test description, the
 * test history and many more.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public interface IDDRunResult extends IDDRunStatusCodes {

	/** State for succesful test execution */
	public final static int STATE_SUCCESSFULL = 1;

	/** State if the test execution has errors */
	public final static int STATE_ERRORS_ON_COMPUTATION = 3;

	/**
	 * Returns the start time of the dd run.
	 * 
	 * @return the start time of the dd run.
	 */
	public Date getStartTime();

	/**
	 * Returns the stop time of the dd run.
	 * 
	 * @return the stop time of the dd run.
	 */
	public Date getStopTime();

	/**
	 * Set the stop time of the dd run.
	 * 
	 * @param stopTime
	 *            the stop time of the dd run.
	 */
	public void setStopTime(Date stopTime);

	/**
	 * Set the delta set including all computed deltas between the passing and
	 * the failing run.
	 * 
	 * @param computedDeltaSet
	 *            all computed deltas between the passing and the failing run.
	 */
	public void setComputedDeltaSet(DeltaSet computedDeltaSet);

	/**
	 * Returns all computed deltas as delta set.
	 * 
	 * @return all computed deltas as delta set.
	 */
	public DeltaSet getComputedDeltaSet();

	/**
	 * Set the minimized deltas computed by the delta debugging algorithm.
	 * 
	 * @param minimizedDeltaSet
	 *            the minimized deltas.
	 */
	public void setMinimizedDeltaSet(DeltaSet minimizedDeltaSet);

	/**
	 * Returns the minimized deltas computed by the delta debugging algorithm.
	 * 
	 * @return the minimized deltas computed by the delta debugging algorithm.
	 */
	public DeltaSet getMinimizedDeltaSet();

	/**
	 * Returns the dd run description.
	 * 
	 * @return the dd run description.
	 */
	public IDDRunDescription getDDRunDescription();

	/**
	 * Returns the name of this dd run result. The name will be used to
	 * represent the dd run result at the ui.
	 * 
	 * @return the name of this dd run result.
	 */
	public String getName();

	/**
	 * Return the state of this dd run result. This can be either successfull,
	 * or error.
	 * 
	 * @return the state of this dd run result.
	 */
	public int getState();

	/**
	 * Add one test result to this dd run result. A test result is the result of
	 * the junit test if we apply a specified delta set.
	 * 
	 * @param testResult
	 *            the test result for one specific junit test.
	 */
	public void addTestResult(IDDTestResult testResult);

	/**
	 * Returns the test results of this dd run result in the order they has been
	 * added to this dd run result.
	 * 
	 * @return the test results of this dd run result.
	 */
	public IDDTestResult[] getTestResults();

	/**
	 * Add the given status to this dd run result.
	 * 
	 * @param status
	 *            the status you want to add to this dd run result.
	 */
	public void addStatus(IStatus status);

	/**
	 * Returns true if at least one status in this dd run result has the
	 * severity <code>IStatus.ERROR</code>.
	 * 
	 * @return if at least one status in this dd run result has the severity
	 *         <code>IStatus.ERROR</code>.
	 */
	public boolean hasErrorStatus();

	/**
	 * Returns true if at least one status in this dd run result has the
	 * severity <code>IStatus.INFO</code>.
	 * 
	 * @return if at least one status in this dd run result has the severity
	 *         <code>IStatus.INFO</code>.
	 */
	public boolean hasInfoStatus();

	/**
	 * Returns a collection of all status that have been added to this dd run
	 * result.
	 * 
	 * @return a collection of all status that have been added to this dd run
	 *         result.
	 */
	public Collection getStatus();

	/**
	 * Returns a collection of all error status that have been added to this dd
	 * run result.
	 * 
	 * @return a collection of all error status that have been added to this dd
	 *         run result.
	 */
	public Collection getErrorStatus();

	/**
	 * Returns a collection of all info status that have been added to this dd
	 * run result.
	 * 
	 * @return a collection of all info status that have been added to this dd
	 *         run result.
	 */
	public Collection getInfoStatus();

	/**
	 * Look if the a status with the given status code was added to this dd run
	 * result. If it was the method return true, otherwise false.
	 * 
	 * @param statusCode
	 *            the status code you want to look for.
	 * @return true if a status with the given status code was added to this dd
	 *         run result, otherwise false.
	 */
	public boolean hasStatus(int statusCode);

}
