/*
 * $Id: MethodTrace.java,v 1.5 2005/08/04 18:54:02 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Jul 19, 2005 1:40:08 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.matcher;

import java.io.Serializable;
import java.util.Arrays;

import org.deltadebugging.ddstate.launcher.server.instrumenter.IdentifierMap;
import org.deltadebugging.ddstate.launcher.server.instrumenter.MethodIdentifier;

/**
 * A method trace is the trace of the called methods of a program excecution.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MethodTrace implements Serializable {

    static final long serialVersionUID = 7045920216421705019L;

    /** The initial size of the method id trace and the method count trace. */
    private final static int INITIAL_TRACE_SIZE = 100;

    /** The initial size of the method count trace map. */
    private final static int INITIAL_CTRACE_MAP_SIZE = 10;

    /**
     * The identifier map maps the method signature to a unique method id and
     * vice versa.
     */
    private final IdentifierMap fIdentifierMap;

    /** The trace is a trace of method id's. */
    private int[] fTrace = null;

    /**
     * The count trace is a trace of the calling count of the methods. If you
     * look at the trace at position i you can found the corresponding calling
     * count of the method in this trace.
     */
    private int[] fCountTrace = null;

    /**
     * The count trace map stores the actual number of method calls for the
     * various methodID. You can look how often a method is called if you get
     * the methodID and read the number from
     * <code>this.fCTraceMap[methodID]</code>.
     */
    private int[] fCTraceMap = null;

    /** The actual number of entries in the method trace */
    private int fActualNumber = 0;

    public MethodTrace(IdentifierMap identifierMap) {
        this.fIdentifierMap = identifierMap;
        this.fTrace = new int[INITIAL_TRACE_SIZE];
        this.fCountTrace = new int[INITIAL_TRACE_SIZE];
        this.fCTraceMap = new int[INITIAL_CTRACE_MAP_SIZE];
    }

    private MethodTrace(IdentifierMap identifierMap, int[] trace,
            int[] countTrace, int[] cTraceMap) {
        this.fIdentifierMap = identifierMap;
        this.fTrace = trace;
        this.fCountTrace = countTrace;
        this.fCTraceMap = cTraceMap;
        this.fActualNumber = this.fTrace.length;
    }

    /**
     * Returns the actual number of entries in the method trace.
     * 
     * @return the actual number of entries in the method trace.
     */
    public int getLength() {
        return this.fActualNumber;
    }

    /**
     * Add the given methoID to the method trace. Call this method if you enter
     * a method.
     * 
     * @param methodID
     *            The id of the method you enter.
     */
    public void addMethodEnter(int methodID) {
        this.addIDToTrace(methodID);
    }

    /**
     * Add the negative number of the given methodID to the method trace. Call
     * this method if you leave a method.
     * 
     * @param methodID
     *            The id of the method you leave.
     */
    public void addMethodExit(int methodID) {
        this.addIDToTrace(-1 * methodID);
    }

    /**
     * Returns the identifier map.
     * 
     * @return the identifier map.
     */
    public IdentifierMap getIdentifierMap() {
        return this.fIdentifierMap;
    }

    /**
     * Returns the method id trace as int array. This includes all method calls
     * and exits.
     * 
     * @return the method id trace as int array.
     */
    public int[] getMethodIDTrace() {
        int[] result = new int[this.fActualNumber];

        System.arraycopy(this.fTrace, 0, result, 0, this.fActualNumber);

        return result;
    }

    /**
     * Returns the method count trace. This trace includes in every entry the
     * number how often the method is count. <br>
     * 
     * E.g. If you look at postion 14 in the method it trace and you found the
     * methodID 9. Now you look at the method count trace on position 14 and you
     * found the entry 3. You can say the the method with the id 9 is called at
     * this position (14) so far 3 times.
     * 
     * @return the method count trace.
     */
    public int[] getMethodCountTrace() {
        int[] result = new int[this.fActualNumber];

        System.arraycopy(this.fCountTrace, 0, result, 0, this.fActualNumber);

        return result;
    }

    /**
     * Add the given methodID to the method trace and compute his count.
     * 
     * @param methodID
     *            the methodID you want to add the method trace.
     */
    private void addIDToTrace(int methodID) {
        if (isTraceLimitReached())
            increaseTraceSize();

        this.fTrace[this.fActualNumber] = methodID;

        if (methodID > 0) {
            this.fCountTrace[this.fActualNumber] = this
                    .increaseCountForMethodID(methodID);
        } else {
            this.fCountTrace[this.fActualNumber] = this
                    .getCountForMethodID(methodID * -1);
        }

        this.fActualNumber++;
    }

    /**
     * Returns true if the trace limit is reached, otherwise false.
     * 
     * @return true if the trace limit is reached, otherwise false.
     */
    private boolean isTraceLimitReached() {
        return this.fTrace.length == this.fActualNumber;
    }

    /**
     * Increase the actual count for the given method and return the new count.
     * 
     * @param methodID
     *            the methodID you want to increase the count.
     * @return the new count for the given methodID.
     */
    private int increaseCountForMethodID(int methodID) {
        if (this.fCTraceMap.length < (methodID + 1)) {
            this.increaseCTraceMapSize(methodID + 1);
        }

        this.fCTraceMap[methodID] = this.fCTraceMap[methodID] + 1;

        return this.fCTraceMap[methodID];
    }

    /**
     * Returns the actual count for the given methodID.
     * 
     * @param methodID
     *            the methodID you want to retrieve the count.
     * @return the actual count for the given methodID.
     */
    private int getCountForMethodID(int methodID) {
        int result = 0;

        if (this.fCTraceMap.length < (methodID + 1)) {
            this.increaseCTraceMapSize(methodID + 1);
        }

        result = this.fCTraceMap[methodID];

        return result;
    }

    /**
     * Increase the size of the count trace map. The new size is at two times of
     * the old size but at least minsize.
     * 
     * @param minsize
     *            the minsize for the increased count trace map.
     */
    private void increaseCTraceMapSize(int minsize) {
        int newsize = 2 * this.fCTraceMap.length;

        while (newsize < minsize) {
            newsize = newsize * 2;
        }

        int[] newCTraceMap = new int[newsize];
        Arrays.fill(newCTraceMap, 0);

        System.arraycopy(this.fCTraceMap, 0, newCTraceMap, 0,
                this.fCTraceMap.length);
        this.fCTraceMap = newCTraceMap;
    }

    /**
     * Increase the trace size and the count trace size. The new size is two
     * times the old size.
     */
    private void increaseTraceSize() {
        // increase the size of the method id trace
        int[] newTrace = new int[2 * this.fTrace.length];

        System.arraycopy(this.fTrace, 0, newTrace, 0, this.fTrace.length);
        this.fTrace = newTrace;

        // increase the size of the method count trace
        int[] newCountTrace = new int[2 * this.fCountTrace.length];

        System.arraycopy(this.fCountTrace, 0, newCountTrace, 0,
                this.fCountTrace.length);
        this.fCountTrace = newCountTrace;
    }

    /**
     * Returns the method id at the given index.
     * 
     * @param index
     *            a number between 0 and <code>this.getLength() - 1</code>.
     * @return the method id at the given index.
     */
    public int getMethodIDAt(int index) {
        if (this.checkIndexOutOfBound(index)) {
            throw new ArrayIndexOutOfBoundsException(
                    "Given index is out of bound");
        }

        return this.fTrace[index];
    }

    /**
     * Returns the sub trace of this method trace start at the given start until
     * the end.
     * 
     * @param startIndex
     * @return a sub trace of the method trace.
     */
    public MethodTrace getSubTrace(int startIndex) {
        return this.getSubTrace(startIndex, this.getLength() - 1);
    }

    /**
     * Returns the sub trace of this method trace start at the given start index
     * and end at the given end index.
     * 
     * @param startIndex
     * @param endIndex
     * @return a sub trace of the method trace.
     */
    public MethodTrace getSubTrace(int startIndex, int endIndex) {
        MethodTrace result = null;

        if (this.checkIndexOutOfBound(startIndex)) {
            throw new ArrayIndexOutOfBoundsException(
                    "Given start index is out of bound");
        }

        if (this.checkIndexOutOfBound(endIndex)) {
            throw new ArrayIndexOutOfBoundsException(
                    "Given end index is out of bound");
        }

        int newLength = (endIndex - startIndex) + 1;
        int[] newTrace = new int[newLength];
        int[] newCountTrace = new int[newLength];
        int[] newCTraceMap = new int[this.fCTraceMap.length];

        System.arraycopy(this.fTrace, startIndex, newTrace, 0, newLength);
        System.arraycopy(this.fCountTrace, startIndex, newCountTrace, 0,
                newLength);
        System.arraycopy(this.fCTraceMap, 0, newCTraceMap, 0,
                this.fCTraceMap.length);

        result = new MethodTrace(this.fIdentifierMap, newTrace,
                newCountTrace, newCTraceMap);

        return result;
    }

    /**
     * Returns true if the given index is out of bound, otherwise false.
     * 
     * @param index
     *            a number between 0 and <code>this.getCount() - 1</code>.
     * @return true if the given index is out of bound, otherwise false.
     */
    private boolean checkIndexOutOfBound(int index) {
        return (index < 0 || index > this.getLength() - 1);
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer result = new StringBuffer();

        for (int i = 0; i < this.fActualNumber; i++) {
            int methodID = this.fTrace[i];

            MethodIdentifier methodIdentifier = null;
            if (methodID > 0) {
                methodIdentifier = this.fIdentifierMap
                        .getMethodIdentifier(methodID);

                if (methodIdentifier != null) {
                    result.append("Method enter: "
                            + methodIdentifier.toString() + " (# "
                            + this.fCountTrace[i] + ")");
                } else {
                    result
                            .append("Error: cannot found method identifier for methodID = "
                                    + methodID);
                }
            } else {
                methodIdentifier = this.fIdentifierMap.getMethodIdentifier(-1
                        * methodID);

                if (methodIdentifier != null) {
                    result.append("Method exit: " + methodIdentifier.toString()
                            + " (# " + this.fCountTrace[i] + ")");
                } else {
                    result
                            .append("Error: cannot found method identifier for methodID = "
                                    + (-1 * methodID));
                }
            }

            if (i < (this.fActualNumber - 1))
                result.append("\n");
        }

        return result.toString();
    }
}
