/*
 * $Id: JUnitTestRunner.java,v 1.35 2006/01/08 10:21:47 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 12.02.2005 11:39:43 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.server;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Enumeration;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestFailure;
import junit.framework.TestResult;
import junit.runner.BaseTestRunner;
import junit.runner.TestSuiteLoader;

import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;
import org.deltadebugging.ddstate.launcher.client.ObjectCreationException;
import org.deltadebugging.ddstate.launcher.matcher.MethodTrace;
import org.deltadebugging.ddstate.launcher.server.instrumenter.CreateInstanceInstrumenter;
import org.deltadebugging.ddstate.launcher.server.instrumenter.IdentifierMap;
import org.deltadebugging.ddstate.launcher.server.instrumenter.MethodTraceWriter;
import org.deltadebugging.ddstate.launcher.server.instrumenter.SimpleMethodTraceInstrumenter;
import org.deltadebugging.ddstate.launcher.server.instrumenter.rtclass.ClassController;
import org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate;
import org.deltadebugging.ddstate.launcher.tester.predicate.PredicateFactory;
import org.eclipse.jdt.internal.junit.runner.JUnitMessages;

/**
 * COMMENTME Comment type DDJunitTestRunner
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class JUnitTestRunner extends BaseTestRunner {

	private static final String STRING_CLASS_NAME = String.class.getName();

	private boolean fDebugMode = false;

	private static final String SET_UP_TEST_METHOD_NAME = "setUpTest";

	private ClassLoader currentClassLoader = null;

	private final static int OBJECT_ARRAY_INIT_SIZE = 10;

	private static Object[] newObjects = new Object[OBJECT_ARRAY_INIT_SIZE];

	private static int actualIndex = 0; // points to the next free element

	private TestSuiteLoader fTestSuiteLoader = null;

	/**
	 * Error messages for the debug output and the exception hanlding
	 */
	private final static String ERR_MSG_CREATE_TEST = "Error on creating test";

	private final static String ERR_MSG_LOAD_STRING_CLASS = "Error on loading class java.lang.String";

	private final static String ERR_MSG_RETR_STRING_CONST = "Error on retrieving constructor for java.lang.String";

	private final static String ERR_MSG_CREATE_STRING = "Error on creating new object of type java.lang.String";

	private final static String ERR_MSG_LOAD_CLASS = "Error on loading class ";

	private final static String ERR_MSG_RETR_CONST = "Error retrieving constructor for class ";

	private final static String ERR_MSG_CREATE_OBJECT = "Error constructing new object of type ";

	private final static String ERR_MSG_UNEXCPECTED_ERROR_ON_TESTEXCEUTION = "Unexcpected error occured on test excecution";

	private final static String ERR_MSG_TEST_RUN_ABORTED = "Test run aborted with errors";

	private void checkNewObjectsSize() {
		int length = newObjects.length;

		if (actualIndex == length) {
			// resize array
			Object[] tmpObjects = new Object[2 * length];
			System.arraycopy(newObjects, 0, tmpObjects, 0, length);
			newObjects = tmpObjects;
		}
	}

	private synchronized int registerNewCreatedObject(Object newObject) {
		this.traceMessage("Try to register new object");
		this.checkNewObjectsSize();

		newObjects[actualIndex] = newObject;
		this.traceMessage("New object succesfull registered");

		return actualIndex++;
	}

	public JUnitTestRunner() {
		this.fDebugMode = false;
	}

	public JUnitTestRunner(boolean debugMode) {
		this.fDebugMode = debugMode;
	}

	public int createNewClass(String instanceType)
			throws ObjectCreationException {
		int result = -1;

		this.traceMessage("Create Class of type " + instanceType);

		try {
			Class c = this.loadClassByName(instanceType);
			result = this.registerNewCreatedObject(c);
		} catch (ClassNotFoundException e) {
			this.traceError(ERR_MSG_LOAD_STRING_CLASS, e);
			throw new ObjectCreationException(instanceType, ERR_MSG_LOAD_CLASS,
					e);
		}

		return result;
	}

	public int createNewString(String string) throws ObjectCreationException {
		int result = -1;

		this.traceMessage("Create String '" + string + "'");

		try {
			Class c = this.currentClassLoader.loadClass(STRING_CLASS_NAME);

			Constructor constr = c.getConstructor(new Class[] { c });
			result = this.registerNewCreatedObject(constr
					.newInstance(new Object[] { string }));

		} catch (ClassNotFoundException e) {
			this.traceError(ERR_MSG_LOAD_STRING_CLASS, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_LOAD_STRING_CLASS, e);

		} catch (SecurityException e) {
			this.traceError(ERR_MSG_RETR_STRING_CONST, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_RETR_STRING_CONST, e);

		} catch (NoSuchMethodException e) {
			this.traceError(ERR_MSG_RETR_STRING_CONST, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_RETR_STRING_CONST, e);

		} catch (IllegalArgumentException e) {
			this.traceError(ERR_MSG_CREATE_STRING, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_CREATE_STRING, e);

		} catch (InstantiationException e) {
			this.traceError(ERR_MSG_CREATE_STRING, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_CREATE_STRING, e);

		} catch (IllegalAccessException e) {
			this.traceError(ERR_MSG_CREATE_STRING, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_CREATE_STRING, e);

		} catch (InvocationTargetException e) {
			this.traceError(ERR_MSG_CREATE_STRING, e);
			throw new ObjectCreationException(STRING_CLASS_NAME,
					ERR_MSG_CREATE_STRING, e);

		}

		return result;
	}

	public int createNewArray(String typeName, int length)
			throws ObjectCreationException {
		int result = -1;

		this.traceMessage("Create Array " + typeName + " with length = "
				+ length);

		if (typeName.endsWith("[]")) {
			typeName = typeName.substring(0, typeName.length() - 2);
		}

		Class c;
		try {
			ClassController.WITH_STATIC_BLOCK = false;
			c = this.loadClassByName(typeName);			

		} catch (ClassNotFoundException e) {
			this.traceError(ERR_MSG_LOAD_CLASS + typeName, e);
			throw new ObjectCreationException(ERR_MSG_LOAD_CLASS + typeName, e);

		} finally {
			ClassController.WITH_STATIC_BLOCK = true;

		}
		
		result = this.registerNewCreatedObject(Array.newInstance(c, length));

		return result;
	}

	public void loadClass(String className) throws ClassNotFoundException {
		this.loadClassByName(className);
	}

	private Class loadClassByName(String typeName)
			throws ClassNotFoundException {
		Class result = null;

		if (typeName.equals("char")) {
			result = char.class;
		} else if (typeName.equals("int")) {
			result = int.class;
		} else if (typeName.equals("long")) {
			result = long.class;
		} else if (typeName.equals("double")) {
			result = double.class;
		} else if (typeName.equals("boolean")) {
			result = boolean.class;
		} else if (typeName.equals("float")) {
			result = float.class;
		} else if (typeName.equals("byte")) {
			result = byte.class;
		} else if (typeName.equals("short")) {
			result = short.class;
		} else {
			result = this.currentClassLoader.loadClass(typeName);
		}

		if (result == null) {
			throw new ClassNotFoundException("Class '" + typeName
					+ "' not found");
		}

		return result;
	}

	public int createNewObject(String className) throws ObjectCreationException {
		int result = -1;

		this.traceMessage("Create Object " + className);

		Class c;
		try {
			ClassController.WITH_STATIC_BLOCK = false;
			c = this.currentClassLoader.loadClass(className);

			Constructor constr = null;

			if (className.equals("java.lang.Object")
					|| className.equals(ClassController.class.getName())) {
				constr = c.getConstructor(new Class[] {});
				result = this.registerNewCreatedObject(constr
						.newInstance(new Object[] {}));

			} else {
				constr = c
						.getConstructor(new Class[] { ClassController.class });
				this.traceMessage("Constructor of " + className + " loaded");
				constr.setAccessible(true);

				result = this
						.registerNewCreatedObject(constr
								.newInstance(new Object[] { ClassController.instance }));

			}
		} catch (ClassNotFoundException e) {
			this.traceError(ERR_MSG_LOAD_CLASS + className, e);
			throw new ObjectCreationException(ERR_MSG_LOAD_CLASS + className, e);

		} catch (SecurityException e) {
			this.traceError(ERR_MSG_RETR_CONST + className, e);
			throw new ObjectCreationException(className, ERR_MSG_RETR_CONST
					+ className, e);

		} catch (NoSuchMethodException e) {
			this.traceError(ERR_MSG_RETR_CONST + className, e);
			throw new ObjectCreationException(className, ERR_MSG_RETR_CONST
					+ className, e);

		} catch (IllegalArgumentException e) {
			this.traceError(ERR_MSG_CREATE_OBJECT + className, e);
			throw new ObjectCreationException(className, ERR_MSG_CREATE_OBJECT
					+ className, e);

		} catch (InstantiationException e) {
			this.traceError(ERR_MSG_CREATE_OBJECT + className, e);
			throw new ObjectCreationException(className, ERR_MSG_CREATE_OBJECT
					+ className, e);

		} catch (IllegalAccessException e) {
			this.traceError(ERR_MSG_CREATE_OBJECT + className, e);
			throw new ObjectCreationException(className, ERR_MSG_CREATE_OBJECT
					+ className, e);

		} catch (InvocationTargetException e) {
			this.traceError(ERR_MSG_CREATE_OBJECT + className, e);
			throw new ObjectCreationException(className, ERR_MSG_CREATE_OBJECT
					+ className, e);

		} finally {

			ClassController.WITH_STATIC_BLOCK = true;
		}

		return result;
	}

	public int getCurrentClassLoaderID() {
		int result = -1;

		if (this.currentClassLoader instanceof ExtTestCaseClassLoader) {
			result = ((ExtTestCaseClassLoader) this.currentClassLoader).getID();
		}

		return result;
	}

	public MethodTrace generateTrace(IJUnitTestDescription testDesc,
			IdentifierMap identifierMap) throws GenerateTraceException {
		MethodTrace result = null;

		this.traceMessage("Start tracing test run '" + testDesc + "'");

		result = new MethodTrace(identifierMap);

		MethodTraceWriter.setTraceContainer(result);
		MethodTraceWriter.enabledTracing();

		SimpleMethodTraceInstrumenter traceInstrumenter = new SimpleMethodTraceInstrumenter(
				identifierMap, MethodTraceWriter.class.getName());

		try {
			this.runTest(testDesc, new ExtTestSuiteLoader(traceInstrumenter));

		} catch (Throwable t) {
			// if an unexpected exception is thrown inform the user about the
			// console
			this.traceError(ERR_MSG_UNEXCPECTED_ERROR_ON_TESTEXCEUTION, t);
			throw new GenerateTraceException("Exception on generating trace", t);

		} finally {
			MethodTraceWriter.disabledTracing();

		}

		this.traceMessage("Finished tracing test run '" + testDesc + "'");

		return result;
	}

	public IPredicate run(IJUnitTestDescription testDesc) {
		IPredicate result = null;

		try {
			TestResult testResult = this.runTest(testDesc,
					new ExtTestSuiteLoader(new CreateInstanceInstrumenter()));

			traceTestResult(testResult);
			result = PredicateFactory.createPredicate(testResult, testDesc);

		} catch (Throwable t) {
			// if an unexpected exception is thrown inform the user about the
			// console
			this.traceError(ERR_MSG_UNEXCPECTED_ERROR_ON_TESTEXCEUTION, t);

			result = PredicateFactory.createPredicate(t, testDesc);
		}

		return result;
	}

	private TestResult runTest(IJUnitTestDescription testDesc,
			TestSuiteLoader testSuiteLoader) throws Throwable {
		TestResult result = new TestResult();

		this.traceMessage("Run Test '" + testDesc + "'");
		this.setTestSuiteLoader(testSuiteLoader);

		try {
			result.addListener(this);

			String suiteClassName = testDesc.getTestCaseClassName();
			String testName = testDesc.getTestName();

			Test test = null;

			// if (this.fRerunTest != null) {
			// Class reloadedTestClass = null;
			//
			// try {
			// reloadedTestClass = getLoader().reload(
			// this.fRerunTest.getClass());
			//
			// } catch (ClassNotFoundException e) {
			// String errMsg = "Error on loading the test class "
			// + this.fRerunTest.getClass() + "'";
			// this.traceError(errMsg, e);
			// }
			//
			// test = TestSuite.createTest(reloadedTestClass, testDesc
			// .getTestName());
			// } else {
			test = this.getTest(suiteClassName, testName);
			// }

			// Remember the current class loader of the test class. This is
			// necessary because if we want to modify static variables of a test
			// and the test is loaded more than one time in the current VM we
			// have to look for the corresponding classloader to identify the
			// test class unique.
			this.currentClassLoader = test.getClass().getClassLoader();

			this.traceMessage("Current class loader ID = "
					+ this.getCurrentClassLoaderID() + "(CL = "
					+ this.currentClassLoader + ")");

			test.run(result);

			this.traceMessage("Test '" + testDesc + "' finished");

		} catch (Throwable t) {
			this.traceError(ERR_MSG_TEST_RUN_ABORTED, t);

			throw t;

		} finally {
			result.removeListener(this);

		}

		return result;
	}

	private Test createTest(String testName, Class testClass) {
		Class[] classArgs = { String.class };
		Test test;
		Constructor constructor = null;
		try {
			try {
				constructor = testClass.getConstructor(classArgs);
				test = (Test) constructor
						.newInstance(new Object[] { testName });
			} catch (NoSuchMethodException e) {
				// try the no arg constructor supported in 3.8.1
				constructor = testClass.getConstructor(new Class[0]);
				test = (Test) constructor.newInstance(new Object[0]);
				if (test instanceof TestCase)
					((TestCase) test).setName(testName);
			}
			if (test != null)
				return test;

		} catch (InstantiationException e) {
			this.traceError(ERR_MSG_CREATE_TEST, e);
		} catch (IllegalAccessException e) {
			this.traceError(ERR_MSG_CREATE_TEST, e);
		} catch (InvocationTargetException e) {
			this.traceError(ERR_MSG_CREATE_TEST, e);
		} catch (NoSuchMethodException e) {
			this.traceError(ERR_MSG_CREATE_TEST, e);
		} catch (ClassCastException e) {
			this.traceError(ERR_MSG_CREATE_TEST, e);
		}
		return warning("Could not create test \'" + testName + "\' "); //$NON-NLS-1$ //$NON-NLS-2$
	}

	private Test getTest(String suiteClassName, String testName) {
		Class testClass = null;
		try {
			testClass = loadSuiteClass(suiteClassName);

		} catch (ClassNotFoundException e) {
			String clazz = e.getMessage();
			if (clazz == null)
				clazz = suiteClassName;

			runFailed(JUnitMessages.getFormattedString(
					"RemoteTestRunner.error.classnotfound", clazz)); //$NON-NLS-1$
			this.traceError("Cannot found class for given suite class name '"
					+ suiteClassName + "'", e);

			return null;

		} catch (Exception e) {
			runFailed(JUnitMessages.getFormattedString(
					"RemoteTestRunner.error.exception", e)); //$NON-NLS-1$
			this.traceError(
					"Unexceptect exception on loading class fr suite class name '"
							+ suiteClassName + "'", e);

			return null;
		}

		Method setup = null;
		Test reloadedTest = this.createTest(testName, testClass);

		try {
			setup = testClass.getMethod(SET_UP_TEST_METHOD_NAME,
					new Class[] { Test.class });
		} catch (SecurityException e1) {
			return reloadedTest;
		} catch (NoSuchMethodException e) {
			return reloadedTest;
		}
		if (setup.getReturnType() != Test.class)
			return warning(JUnitMessages
					.getString("RemoteTestRunner.error.notestreturn")); //$NON-NLS-1$
		if (!Modifier.isPublic(setup.getModifiers()))
			return warning(JUnitMessages
					.getString("RemoteTestRunner.error.shouldbepublic")); //$NON-NLS-1$
		if (!Modifier.isStatic(setup.getModifiers()))
			return warning(JUnitMessages
					.getString("RemoteTestRunner.error.shouldbestatic")); //$NON-NLS-1$
		try {
			Test test = (Test) setup
					.invoke(null, new Object[] { reloadedTest });
			if (test == null)
				return warning(JUnitMessages
						.getString("RemoteTestRunner.error.nullreturn")); //$NON-NLS-1$
			return test;
		} catch (IllegalArgumentException e) {
			return warning(JUnitMessages.getFormattedString(
					"RemoteTestRunner.error.couldnotinvoke", e)); //$NON-NLS-1$
		} catch (IllegalAccessException e) {
			return warning(JUnitMessages.getFormattedString(
					"RemoteTestRunner.error.couldnotinvoke", e)); //$NON-NLS-1$
		} catch (InvocationTargetException e) {
			return warning(JUnitMessages
					.getFormattedString(
							"RemoteTestRunner.error.invocationexception", e.getTargetException())); //$NON-NLS-1$
		}
	}

	private void traceTestResult(TestResult testResult) {
		this.traceMessage("Failure Count: " + testResult.failureCount());
		this.traceMessage("Error Count: " + testResult.errorCount());

		this.traceMessage("Result:");

		if (testResult.errorCount() > 0) {
			if (this.fDebugMode) {
				this.traceMessage("\t-> FAIL <-");
				int i = 1;

				Enumeration errors = testResult.errors();

				this.traceMessage("Errors:");
				while (errors.hasMoreElements()) {
					TestFailure error = (TestFailure) errors.nextElement();
					this.traceMessage("\t" + i++ + ".:"
							+ error.exceptionMessage());
					this.traceMessage(error.trace());
				}
			}

		} else if (testResult.failureCount() > 0) {
			if (this.fDebugMode) {
				this.traceMessage("\t-> FAIL <-");
				int i = 1;

				Enumeration failures = testResult.failures();

				this.traceMessage("Failures");
				while (failures.hasMoreElements()) {
					TestFailure failure = (TestFailure) failures.nextElement();
					this.traceMessage("\t" + i++ + ".:"
							+ failure.exceptionMessage());
					this.traceMessage(failure.trace());
				}
			}

		} else {
			this.traceMessage("\t-> PASS <-");
		}
		this.traceMessage("------------------------------------");
	}

	public TestSuiteLoader getLoader() {
		return this.fTestSuiteLoader;
	}

	private void setTestSuiteLoader(TestSuiteLoader testSuiteLoader) {
		this.fTestSuiteLoader = testSuiteLoader;
	}

	/**
	 * Trace the given message if debug is enabled.
	 * 
	 * @param message
	 *            message to trace.
	 */
	public void traceMessage(String message) {
		if (this.fDebugMode) {
			System.out.println(message);
		}
	}

	/**
	 * Trace the given message and the given exception if debug is enabled.
	 * 
	 * @param message
	 *            message to trace.
	 * @param exception
	 *            exception to trace.
	 */
	public void traceError(String message, Throwable exception) {
		if (this.fDebugMode) {
			System.err.println(message);
			exception.printStackTrace();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.runner.BaseTestRunner#testStarted(java.lang.String)
	 */
	public void testStarted(String testName) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.runner.BaseTestRunner#testEnded(java.lang.String)
	 */
	public void testEnded(String testName) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.runner.BaseTestRunner#testFailed(int, junit.framework.Test,
	 *      java.lang.Throwable)
	 */
	public void testFailed(int status, Test test, Throwable t) {
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see junit.runner.BaseTestRunner#runFailed(java.lang.String)
	 */
	protected void runFailed(String message) {
		// do nothing
	}

	/**
	 * Returns a test which will fail and log a warning message.
	 */
	private Test warning(final String message) {
		return new TestCase("warning") { //$NON-NLS-1$
			protected void runTest() {
				fail(message);
			}
		};
	}
}