/*
 * $Id: RemoteTestManager.java,v 1.3 2005/07/21 12:50:28 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Jan 7, 2005 1:10:38 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.server;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

/**
 * COMMENTME Comment type DDRemoteTestManager
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class RemoteTestManager extends UnicastRemoteObject implements
        IRemoteTestManager {

    static final long serialVersionUID = 542405940276038838L;
    
    private String fRMIHost = null;
    private int fRMIPort = -1;
    private boolean fDebugMode = false;
    private Collection fTestRunner = new Vector();

    /**
     * Construct a new testmanager without registring in the jndi context.
     */
    public RemoteTestManager() throws RemoteException {
        super();
    }

    /**
     * Construct a new testmanager and register it on the jndicontext. The
     * jndicontext run on the given port on the localhost.
     * 
     * @param rmiPort
     *            the port on which the registry accepts requests.
     * @param debug
     *            if true the debug output is enabled, otherwise disabled.
     * @throws RemoteException
     *             if failed to bind this manager in a local rmi registry.
     */
    public RemoteTestManager(int rmiPort, boolean debug) throws RemoteException {
        super();
        this.fRMIHost = "localhost";
        this.fRMIPort = rmiPort;
        this.fDebugMode = debug;
        this.registerJNDI();
    }

    /**
     * Constructs an new testmanager and register it on the jndicontext.
     * 
     * @param args
     *            arguments of the main method
     */
    public RemoteTestManager(String[] args) throws RemoteException {
        super();
        this.init(args);
        this.registerJNDI();
    }

    /**
     * @param args
     */
    private void init(String[] args) {
        this.fRMIHost = "localhost";
        this.fRMIPort = 1099;
        this.fDebugMode = false;

        if (args.length % 2 == 0) {
            for (int i = 0; i < args.length; i += 2) {
                if (args[i].toLowerCase().equals("-port")) {
                    this.fRMIPort = new Integer(args[i + 1]).intValue();
                } else if (args[i].toLowerCase().equals("-debug")) {
                    this.fDebugMode = args[3].equals("on") ? true : false;
                }
            }
        }
    }

    public IRemoteTestRunner getTestRunner(boolean debugMode)
            throws RemoteException {
        IRemoteTestRunner result = new RemoteTestRunner(debugMode);
        this.fTestRunner.add(result);

        return result;
    }

    /**
     * Register this testmanager in the jndicontext.
     */
    private void registerJNDI() throws RemoteException {

        // If localhost start registry on given port
        if (this.fRMIHost.equals("localhost")
                || this.fRMIHost.equals("127.0.0.1")) {
            this.traceMessage("Start RMIServer on " + this.fRMIHost
                    + " (Port = " + this.fRMIPort + ") ... ");
            try {
                LocateRegistry.createRegistry(this.fRMIPort);
                this.traceMessage("Start ... OK");
            } catch (RemoteException e) {
                this.traceMessage("Start ... FAIL");
                throw e;
            }
        }

        // Bind this DDRemoteTestManager
        String url = RemoteTestManager.createURL(this.fRMIHost, this.fRMIPort);

        this.traceMessage("Bind DDRemoteTestManager on URL '" + url + "' ... ");
        try {
            Naming.rebind(url, this);
            this.traceMessage("Bind ... OK");
        } catch (RemoteException e) {
            this.traceMessage("Bind ... FAIL");
            throw e;
        } catch (MalformedURLException e) {
            this.traceMessage("Bind ... FAIL");
            throw new RemoteException(
                    "Cannot bind DDRemoteTesetManager on URL '" + url + "'");
        }

    }

    /**
     * Shutdown this testmanager.
     */
    public void shutDown() throws RemoteException {
        int exitCode = 0;

        this.traceMessage("Shutdown TestManager ... ");

        if (this.fRMIHost != null && this.fRMIPort > 0) {
            String url = RemoteTestManager.createURL(this.fRMIHost,
                    this.fRMIPort);

            try {
                Naming.unbind(url);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        this.traceMessage("Shutdown ... OK");

        System.exit(exitCode);
    }

    /**
     * Create a unique URL for the JNDI-Context. This url can be used to bind,
     * rebind, unbind or lookup the DDRemoteTestManager in the rmiregistry of
     * the given host on the given port.
     * 
     * @param rmiHost
     *            host where the rmiregistry run.
     * @param rmiPort
     *            port where the rmiregistry listen.
     * @return unique url
     */
    public static String createURL(String rmiHost, int rmiPort) {
        String result = "rmi://" + rmiHost + ":" + rmiPort + "/"
                + IRemoteTestManager.SERVICE_NAME;
        return result;
    }

    /**
     * Trace the given message if debug is enabled.
     * 
     * @param message
     *            message to trace.
     */
    public void traceMessage(String message) {
        if (this.fDebugMode) {
            System.out.println(message);
        }
    }

    // Start the DDRemoteTestManager with the given arguments
    // p.e.:
    //  1) DDRemoteTestManager [-port 1456] [-debug on|off]
    public static void main(String[] args) {
        try {
            new RemoteTestManager(args);
        } catch (RemoteException e) {
            e.printStackTrace();
            System.exit(1);
        } catch (Throwable t) {
            t.printStackTrace();
            System.exit(2);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.server.IRemoteTestManager#hasRunningTests()
     */
    public boolean hasRunningTests() throws RemoteException {
        boolean result = false;

        Iterator itTestRunner = this.fTestRunner.iterator();

        while (itTestRunner.hasNext() && result == false) {
            IRemoteTestRunner testRunner = (IRemoteTestRunner) itTestRunner
                    .next();
            result |= testRunner.isRunning();
        }

        return result;
    }
}