/*
 * $Id: SimpleOutcomeCache.java,v 1.4 2005/10/10 13:35:33 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University Prof. Dr. Andreas Zeller
 * 
 * All rights reserved. This program and the accompanying materials are made
 * available under the terms of the Common Public License v1.0 which accompanies
 * this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Created on 16.07.2005 13:32:25 by @author Karsten Lehmann Mail to:
 * lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.tester;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.deltadebugging.ddcore.DD;
import org.deltadebugging.ddcore.DeltaSet;

/**
 * A simple outcome cache for the delta set. This cache the outcome of a delta
 * debugging test for a given delta set.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
class SimpleOutcomeCache {

    /** The cache for the outcome */
    private Map fCache = new HashMap();

    /**
     * Add the given delta set and test outcome to the cache if it doesn't
     * contain.
     * 
     * @param c
     *            the delta set you want to add to the cache.
     * @param testOutcome
     *            the test outcome of this delta set.
     */
    public void addTestOutcome(DeltaSet c, int testOutcome) {
        if (!this.fCache.containsKey(new DeltaSetWrapper(c))) {
            this.fCache.put(new DeltaSetWrapper(c), new Integer(testOutcome));
        }
    }

    /**
     * Check if the outcome of the given delta set is already cache. If it do so
     * return true, otherwise false.
     * 
     * @param c
     *            the deltaset you look for.
     * @return true if the test outcome of the given delta set is in cache,
     *         otherwise false.
     */
    public boolean existTestOutcome(DeltaSet c) {
        boolean result = false;

        result = this.fCache.containsKey(new DeltaSetWrapper(c));

        return result;
    }

    /**
     * Returns the cached test outcome of the given delta set.
     * 
     * @param c
     *            delta set you want to get the test outcome.
     * @throws IllegalArgumentException
     *             if the test outcome of the given delta set is not in cache.
     * @return the cached test outcome of the given delta set.
     */
    public int getTestOutcome(DeltaSet c) {
        int result = DD.UNRESOLVED;

        if (!this.existTestOutcome(c))
            throw new IllegalArgumentException(
                    "No outcome exitst for DeltaSet\n" + c);

        result = ((Integer) this.fCache.get(new DeltaSetWrapper(c))).intValue();

        return result;
    }

    /**
     * This Wrapper class implements a hash code function and the equals
     * functionality for the delta set. So it is possible to cache the delta
     * sets.
     * 
     * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
     */
    class DeltaSetWrapper {

        /** The delta set of this wrapper class */
        private DeltaSet fDeltaSet = null;

        /**
         * Constructs a new wrapper with the given delta set. You can use the
         * wrapper to cache the delta set.
         * 
         * @param deltaSet
         *            the delta set you want to cache.
         */
        public DeltaSetWrapper(DeltaSet deltaSet) {
            this.fDeltaSet = deltaSet;
        }

        /**
         * Returns the hash code value for this delta set.
         * <p>
         * 
         * This implementation uses exactly the code that is used to define the
         * list hash function in the documentation for the
         * <tt>List.hashCode</tt> method.
         * 
         * @return the hash code value for this delta set.
         */
        public int hashCode() {
            int hashCode = 1;
            Iterator i = this.fDeltaSet.iterator();
            while (i.hasNext()) {
                Object obj = i.next();
                hashCode = 31 * hashCode + (obj == null ? 0 : obj.hashCode());
            }
            return hashCode;
        }

        public boolean equals(Object obj) {
            boolean result = false;

            if (obj instanceof DeltaSetWrapper) {
                DeltaSetWrapper compareObj = (DeltaSetWrapper) obj;

                if (compareObj.getDeltaSet().size() == this.fDeltaSet.size()) {
                    result = compareObj.getDeltaSet().containsAll(
                            this.fDeltaSet);
                }
            }

            return result;
        }

        /**
         * Returns the delta set of this wrapper class.
         * 
         * @return the delta set of this wrapper class.
         */
        public DeltaSet getDeltaSet() {
            return this.fDeltaSet;
        }
    }
}