/*
 * $Id: FailurePredicate.java,v 1.4 2005/09/21 08:45:11 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Sep 12, 2005 11:18:13 AM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.tester.predicate;

import junit.framework.TestFailure;

import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;

/**
 * The failure predicate defines a predicate for every failed junit test run.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
 */
public abstract class FailurePredicate extends Predicate {

	/**
	 * The exception raised on the junit test run. This can also be an
	 * AssertionException if an assert statement fails (only junit asserts).
	 * <p>
	 * We cannont remember the original exception because this can changed by
	 * the classloader.
	 */
	private final String fExceptionName;

	private final String fExceptionMessage;

	/**
	 * The filtered stacktrace of the exception. We filter the stacktrace,
	 * because we are only interested in all method calls above the first test
	 * method call.
	 */
	private final StackTraceElement[] fExcTrace;

	public FailurePredicate(TestFailure failure, IJUnitTestDescription testDesc) {
		super(failure.isFailure() ? STATE_TEST_WITH_FAILURES
				: STATE_TEST_WITH_ERRORS);

		Throwable exception = failure.thrownException();

		this.fExceptionName = exception.getClass().getName();
		this.fExceptionMessage = exception.getMessage();
		this.fExcTrace = this.filterStackTrace(exception.getStackTrace(),
				testDesc.getTestCaseClassName(), testDesc.getTestName());
	}

	protected FailurePredicate(Throwable exception,
			IJUnitTestDescription testDesc) {
		super(STATE_TEST_WITH_ERRORS);

		this.fExceptionName = exception.getClass().getName();
		this.fExceptionMessage = exception.getMessage();
		this.fExcTrace = this.filterStackTrace(exception.getStackTrace(),
				testDesc.getTestCaseClassName(), testDesc.getTestName());
	}

	/**
	 * Returns the full quallified class name of the exception raised on the
	 * junit test run execution.
	 * 
	 * @return the full quallified class name of the exception raised on the
	 *         junit test run execution.
	 */
	public String getExceptionName() {
		return this.fExceptionName;
	}

	/**
	 * Returns the exception message.
	 * 
	 * @return the exception message.
	 */
	public String getExceptionMessage() {
		return this.fExceptionMessage;
	}

	/**
	 * Returns the filtered exception stack trace.
	 * 
	 * @return the filtered exception stack trace.
	 */
	public StackTraceElement[] getExceptionTrace() {
		return this.fExcTrace;
	}

	/**
	 * Returns the filtered exception stacktrace as String.
	 * 
	 * @return the filtered exception stacktrace as String.
	 */
	public String getExceptionTraceAsString() {
		StringBuffer result = new StringBuffer();

		for (int i = 0; i < this.fExcTrace.length; i++) {
			StackTraceElement element = this.fExcTrace[i];
			result.append(element.toString());

			if (i < this.fExcTrace.length - 1)
				result.append("\n");
		}

		return result.toString();
	}

	/**
	 * Filter the given stacktrace from botton until the first occurance of the
	 * given method name in the given class.
	 * 
	 * @param stackTrace
	 *            the stacktrace you want to filter.
	 * @param className
	 *            the name of the class the method call occure.
	 * @param methodName
	 *            the name of the method.
	 * @return the filtered stacktrace.
	 */
	private StackTraceElement[] filterStackTrace(
			StackTraceElement[] stackTrace, String className, String methodName) {
		StackTraceElement[] result = new StackTraceElement[0];

		for (int i = stackTrace.length - 1; i >= 0; i--) {
			StackTraceElement element = stackTrace[i];

			if (element.getClassName().equals(className)
					&& element.getMethodName().equals(methodName)) {
				result = new StackTraceElement[i + 1];
				System.arraycopy(stackTrace, 0, result, 0, i + 1);
				break;
			}

		}

		return result;
	}

	/**
	 * Check if the given stacktrace equals this stacktrace expect the method on
	 * the bottom. We wan't compare the test methods which are not equal in the
	 * passing and the failing run therefor we disregard the first stacktrace
	 * entry.
	 * 
	 * @param stackTrace
	 * @return true if the given stacktrace equals this stacktrace, otherwise
	 *         false.
	 */
	private boolean matchesStackTrace(StackTraceElement[] stackTrace) {
		boolean result = false;

		if (this.fExcTrace.length == stackTrace.length) {

			result = true;

			for (int i = 1; i < stackTrace.length && result; i++) {
				result &= this.fExcTrace[i].getClassName().equals(
						stackTrace[i].getClassName());
				result &= this.fExcTrace[i].getMethodName().equals(
						stackTrace[i].getMethodName());
			}

		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate#matches(org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate)
	 */
	public boolean matches(IPredicate predicate) {
		boolean result = false;

		if (predicate instanceof FailurePredicate) {
			result = this.matches((FailurePredicate) predicate);
		}

		return result;
	}

	/**
	 * Returns true if the given predicate is equal to this predicate. Two
	 * FailurePredicate's are equal if they have the same exception (compare on
	 * class name) and the same filtered stacktrace.
	 * 
	 * @param predicate
	 *            the predicate to compare.
	 * @return true if the given predicate is equal to this predicate, otherwise
	 *         false.
	 */
	public boolean matches(FailurePredicate predicate) {
		boolean result = true;

		result &= this.getExceptionName().equals(predicate.getExceptionName());

		if (result) {
			result &= this.matchesStackTrace(predicate.getExceptionTrace());
		}

		return true;
	}

}
