/*
 * $Id: DDRunResultManager.java,v 1.2 2005/04/25 06:31:57 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 09.04.2005 17:26:16 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.ui.resultmanager;

import java.util.ArrayList;
import java.util.List;

import org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.ui.DDStateUIPlugin;
import org.deltadebugging.ddstate.ui.views.IDDRunResultView;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.util.ListenerList;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;

/**
 * COMMENTME Comment type DDRunResultManager
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DDRunResultManager implements IDDRunResultManager {

    // Personal logger for this class
    final static DDLogger logger = DDStateUIPlugin
            .getLogger(DDRunResultManager.class);

    /**
     * DD run listeners
     */
    ListenerList fListeners = null;

    /**
     * List of registered dd run results
     */
    private List fDDRuns = new ArrayList(10);

    // change notification constants
    private final static int ADDED = 1;
    private final static int REMOVED = 2;

    /**
     * Notifies a dd run result listener of additions or removals
     */
    class DDRunResultNotifier implements ISafeRunnable {

        private IDDRunResultListener fListener;
        private int fType;
        private IDDRunResult[] fChanged;

        /*
         * (non-Javadoc)
         * 
         * @see org.eclipse.core.runtime.ISafeRunnable#handleException(java.lang.Throwable)
         */
        public void handleException(Throwable exception) {
            logger
                    .log(
                            IStatus.ERROR,
                            "Message occured on notification the dd run result listener", //$NON-NLS-1$
                            exception);
        }

        /*
         * (non-Javadoc)
         * 
         * @see org.eclipse.core.runtime.ISafeRunnable#run()
         */
        public void run() throws Exception {
            switch (this.fType) {
            case ADDED:
                this.fListener.ddRunResultsAdded(this.fChanged);
                break;
            case REMOVED:
                this.fListener.ddRunResultsRemoved(this.fChanged);
                break;
            }
        }

        /**
         * Notifies the given listener of the adds/removes
         * 
         * @param results
         *            the dd run results that changed
         * @param update
         *            the type of change
         */
        public void notify(IDDRunResult[] results, int update) {
            if (DDRunResultManager.this.fListeners == null) {
                return;
            }
            this.fChanged = results;
            this.fType = update;
            Object[] copiedListeners = DDRunResultManager.this.fListeners
                    .getListeners();
            for (int i = 0; i < copiedListeners.length; i++) {
                this.fListener = (IDDRunResultListener) copiedListeners[i];
                Platform.run(this);
            }
            this.fChanged = null;
            this.fListener = null;
        }
    }

    /**
     * Fires notification.
     * 
     * @param results
     *            dd run results added/removed
     * @param type
     *            ADD or REMOVE
     */
    private void fireUpdate(IDDRunResult[] results, int type) {
        new DDRunResultNotifier().notify(results, type);
    }

    /**
     * Returns whether the given dd run result view should be brought to the
     * top. The view should not be brought to the top if the view is pinned on a
     * console other than the given console.
     */
    boolean shouldBringToTop(IDDRunResult result, IViewPart resultView) {
        boolean bringToTop = true;
        if (resultView instanceof IDDRunResultView) {
            IDDRunResultView rView = (IDDRunResultView) resultView;
            if (rView.isPinned()) {
                IDDRunResult pinnedDDRunResult = rView.getDDRunResult();
                bringToTop = result.equals(pinnedDDRunResult);
            }
        }
        return bringToTop;
    }

    //    /*
    //     * (non-Javadoc)
    //     *
    //     * @see
    // org.eclipse.ui.console.IConsoleManager#warnOfContentChange(org.eclipse.ui.console.IConsole)
    //     */
    //    public void warnOfContentChange(IConsole console) {
    //        IWorkbenchWindow window = PlatformUI.getWorkbench()
    //                .getActiveWorkbenchWindow();
    //        if (window != null) {
    //            IWorkbenchPage page = window.getActivePage();
    //            if (page != null) {
    //                IConsoleView consoleView = (IConsoleView) page
    //                        .findView(IConsoleConstants.ID_CONSOLE_VIEW);
    //                if (consoleView != null) {
    //                    consoleView.warnOfContentChange(console);
    //                }
    //            }
    //        }
    //    }

    public DDRunResultManager() {
        super();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#addDDRunResultListener(org.deltadebugging.ddstate.ui.views.IDDRunResultListener)
     */
    public void addDDRunResultListener(IDDRunResultListener listener) {
        if (this.fListeners == null) {
            this.fListeners = new ListenerList(5);
        }
        this.fListeners.add(listener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#removeDDRunResultListener(org.deltadebugging.ddstate.ui.views.IDDRunResultListener)
     */
    public void removeDDRunResultListener(IDDRunResultListener listener) {
        if (this.fListeners != null) {
            this.fListeners.remove(listener);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#addDDRunResults(org.deltadebugging.ddstate.ui.views.IDDRunResult[])
     */
    public void addDDRunResults(IDDRunResult[] results) {
        List added = new ArrayList(results.length);
        for (int i = 0; i < results.length; i++) {
            IDDRunResult ddRunResult = results[i];
            if (!this.fDDRuns.contains(ddRunResult)) {
                this.fDDRuns.add(ddRunResult);
                added.add(ddRunResult);
            }
        }
        if (!added.isEmpty()) {
            fireUpdate((IDDRunResult[]) added.toArray(new IDDRunResult[added
                    .size()]), ADDED);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#removeDDRunResults(org.deltadebugging.ddstate.ui.views.IDDRunResult[])
     */
    public void removeDDRunResults(IDDRunResult[] results) {
        List removed = new ArrayList(results.length);
        for (int i = 0; i < results.length; i++) {
            IDDRunResult console = results[i];
            if (this.fDDRuns.remove(console)) {
                removed.add(console);
            }
        }
        if (!removed.isEmpty()) {
            fireUpdate((IDDRunResult[]) removed
                    .toArray(new IDDRunResult[removed.size()]), REMOVED);
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#showDDRunResultView(org.deltadebugging.ddstate.ui.views.IDDRunResult)
     */
    public void showDDRunResultView(final IDDRunResult result) {
        DDStateUIPlugin.getStandardDisplay().asyncExec(new Runnable() {
            public void run() {
                IWorkbenchWindow window = PlatformUI.getWorkbench()
                        .getActiveWorkbenchWindow();
                if (window != null) {
                    IWorkbenchPage page = window.getActivePage();
                    if (page != null) {
                        IViewPart resultView = page
                                .findView(DDStateUIPlugin.DDRUN_RESULT_VIEW_ID);
                        if (resultView == null) {
                            try {
                                resultView = page.showView(
                                        DDStateUIPlugin.DDRUN_RESULT_VIEW_ID,
                                        null, IWorkbenchPage.VIEW_CREATE);
                            } catch (PartInitException pie) {
                                logger
                                        .log(
                                                IStatus.ERROR,
                                                "Error on opening a dd run result view", //$NON-NLS-1$
                                                pie);
                            }
                        }

                        // else {
                        boolean bringToTop = shouldBringToTop(result,
                                resultView);
                        if (bringToTop) {
                            page.bringToTop(resultView);
                        }
                        //                        }
                        if (resultView instanceof IDDRunResultView) {
                            ((IDDRunResultView) resultView).display(result);
                        }
                    }
                }
            }
        });
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.ui.IDDRunResultManager#getDDRunResults()
     */
    public IDDRunResult[] getDDRunResults() {
        return (IDDRunResult[]) this.fDDRuns
                .toArray(new IDDRunResult[this.fDDRuns.size()]);
    }
}