/*
 * $Id: MemoryAccessUtil.java,v 1.30 2005/11/21 17:03:41 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 15.02.2005 15:04:39 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.extractor;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.extractor.exceptions.MemoryAccessException;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.ILocalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.type.IGlobalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.type.ITypeEdge;
import org.deltadebugging.ddstate.memorygraph.util.Path;
import org.deltadebugging.ddstate.memorygraph.value.IArrayElementEdge;
import org.deltadebugging.ddstate.memorygraph.value.IObjectFieldEdge;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IStackFrame;
import org.eclipse.debug.core.model.IVariable;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.debug.core.IJavaArray;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaObject;
import org.eclipse.jdt.debug.core.IJavaReferenceType;
import org.eclipse.jdt.debug.core.IJavaStackFrame;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaValue;
import org.eclipse.jdt.debug.core.IJavaVariable;
import org.eclipse.jdt.internal.debug.core.model.JDIArrayType;
import org.eclipse.jdt.internal.debug.core.model.JDIClassType;
import org.eclipse.jdt.internal.debug.core.model.JDIDebugTarget;
import org.eclipse.jdt.internal.debug.core.model.JDIInterfaceType;
import org.eclipse.jdt.internal.debug.core.model.JDINullValue;

import com.sun.jdi.ArrayType;
import com.sun.jdi.ClassType;
import com.sun.jdi.Field;
import com.sun.jdi.InterfaceType;
import com.sun.jdi.ObjectCollectedException;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.Value;
import com.sun.jdi.VirtualMachine;

/**
 * COMMENTME Comment type MemoryAccessUtil
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MemoryAccessUtil {

	/**
	 * Default values for the existing java primitive types and object
	 * references.
	 * 
	 * @see Java Language Specification (Second Edition) Chapter 4.5.5
	 */
	public final static byte DEFAULT_BYTE_VALUE = (byte) 0;

	public final static short DEFAULT_SHORT_VALUE = (short) 0;

	public final static int DEFAULT_INT_VALUE = 0;

	public final static long DEFAULT_LONG_VALUE = 0L;

	public final static float DEFAULT_FLOAT_VALUE = 0.0f;

	public final static double DEFAULT_DOUBLE_VALUE = 0.0d;

	public final static char DEFAULT_CHAR_VALUE = '\u0000';

	public final static boolean DEFAULT_BOOLEAN_VALUE = false;

	public final static Object DEFAULT_REFERENCE_VALUE = null;

	// Personal logger for this class
	final static DDLogger logger = DDStatePlugin
			.getLogger(MemoryAccessUtil.class);

	private MemoryAccessUtil() {
		// util class is not initializable
	}

	public static IJavaValue generateDefaultValueForVariable(
			IJavaThread thread, IJavaVariable variable)
			throws MemoryAccessException {
		IJavaValue result = null;

		IJavaValue value = MemoryAccessUtil.getJavaValue(variable);
		String signature = null;

		try {
			if (value instanceof JDINullValue) {
				signature = variable.getJavaType().getSignature();

			} else {
				signature = value.getJavaType().getSignature();

			}
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Cannot receive the java type signature", e);
		}

		IJavaDebugTarget jDebugTarget = (IJavaDebugTarget) thread
				.getDebugTarget().getAdapter(IJavaDebugTarget.class);

		switch (signature.charAt(0)) {
		case Signature.C_BOOLEAN:
			result = jDebugTarget.newValue(DEFAULT_BOOLEAN_VALUE);
			break;
		case Signature.C_BYTE:
			result = jDebugTarget.newValue(DEFAULT_BYTE_VALUE);
			break;
		case Signature.C_CHAR:
			result = jDebugTarget.newValue(DEFAULT_CHAR_VALUE);
			break;
		case Signature.C_SHORT:
			result = jDebugTarget.newValue(DEFAULT_SHORT_VALUE);
			break;
		case Signature.C_INT:
			result = jDebugTarget.newValue(DEFAULT_INT_VALUE);
			break;
		case Signature.C_LONG:
			result = jDebugTarget.newValue(DEFAULT_LONG_VALUE);
			break;
		case Signature.C_FLOAT:
			result = jDebugTarget.newValue(DEFAULT_FLOAT_VALUE);
			break;
		case Signature.C_DOUBLE:
			result = jDebugTarget.newValue(DEFAULT_DOUBLE_VALUE);
			break;
		case Signature.C_RESOLVED:
			result = jDebugTarget.nullValue();
			break;
		case Signature.C_ARRAY:
			result = jDebugTarget.nullValue();
			break;
		default:
			throw new RuntimeException("Unexpected signature '" + signature
					+ "'");
		}

		return result;
	}

	private static IJavaValue getJavaValue(IJavaVariable variable)
			throws MemoryAccessException {
		IJavaValue result = null;

		try {
			result = (IJavaValue) variable.getValue().getAdapter(
					IJavaValue.class);
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Cannot acces the value of the variable");
		}

		return result;
	}

	public static List getAllLoadedClasses(IJavaThread thread)
			throws MemoryAccessException {
		JDIDebugTarget debugTarget = MemoryAccessUtil.getDebugTarget(thread);

		return MemoryAccessUtil.getAllLoadedClasses(debugTarget);
	}

	public static List getLoadedClassesByName(IJavaThread thread,
			String className) throws MemoryAccessException {
		JDIDebugTarget debugTarget = MemoryAccessUtil.getDebugTarget(thread);

		return MemoryAccessUtil.getLoadedClassesByName(debugTarget, className);
	}

	private static List getAllLoadedClasses(JDIDebugTarget debugTarget) {
		List result = null;

		VirtualMachine vm = debugTarget.getVM();

		if (vm != null) {
			result = vm.allClasses();
		}

		return result;
	}

	private static List getLoadedClassesByName(JDIDebugTarget debugTarget,
			String className) {
		List result = null;

		VirtualMachine vm = debugTarget.getVM();

		if (vm != null) {
			result = vm.classesByName(className);
		}

		return result;
	}

	public static IJavaStackFrame[] getAllStackframes(IJavaThread thread)
			throws DebugException {
		assert thread != null;
		assert thread.isSuspended();

		IStackFrame[] allFrames = thread.getStackFrames();

		Collection result = new Vector();

		for (int i = 0; i < allFrames.length; i++) {
			IJavaStackFrame frame = (IJavaStackFrame) allFrames[i]
					.getAdapter(IJavaStackFrame.class);

			if (frame != null) {
				result.add(frame);
			} else {
				logger.log(IStatus.WARNING, "Frame of type "
						+ allFrames[i].getClass().getName()
						+ " is not adaptable to IJavaStackFrame.");
			}
		}

		return (IJavaStackFrame[]) result.toArray(new IJavaStackFrame[result
				.size()]);
	}

	public static IJavaVariable[] getAllJavaVariables(IJavaStackFrame stackFrame)
			throws DebugException {
		assert stackFrame != null;

		boolean isNative = false;
		IVariable[] variables = null;

		isNative = stackFrame.isNative();
		if (!isNative) {
			variables = stackFrame.getVariables();
		}

		Collection result = new Vector();

		if (!isNative) {
			for (int i = 0; i < variables.length; i++) {
				IJavaVariable variable = (IJavaVariable) variables[i]
						.getAdapter(IJavaVariable.class);

				if (variable != null) {
					result.add(variable);
				} else {
					logger.log(IStatus.WARNING, "Variable of type "
							+ variables[i].getClass().getName()
							+ " is not adaptable to IJavaVariable.");
				}
			}
		}

		return (IJavaVariable[]) result
				.toArray(new IJavaVariable[result.size()]);
	}

	public static IJavaReferenceType[] getTypesWithStaticVariables(
			IJavaThread thread) throws MemoryAccessException {
		assert thread != null;
		assert thread.isSuspended();

		Collection result = new Vector();

		JDIDebugTarget debugTarget = MemoryAccessUtil.getDebugTarget(thread);
		List classes = MemoryAccessUtil.getAllLoadedClasses(debugTarget);

		Map registeredTypes = new HashMap();

		for (Iterator itClass = classes.iterator(); itClass.hasNext();) {
			ReferenceType refType = (ReferenceType) itClass.next();

			Iterator itFields;
			boolean isRelevantType = false;
			try {
				itFields = refType.fields().iterator();
				
				while (!isRelevantType && itFields.hasNext()) {
					Field field = (Field) itFields.next();

					if (field.isStatic()) { // && !field.isFinal()) {
						isRelevantType = true;
					}
				}

				if (isRelevantType) {
					String refTypeName = refType.name();

					// FIXME look if we have to use the class loader id??
					if (!registeredTypes.containsKey(refTypeName)) {
						registeredTypes.put(refTypeName, refType);
						IJavaReferenceType type = MemoryAccessUtil
								.createNewReferenceType(refType, debugTarget);
						result.add(type);

					} else {
						logger.trace(DDStatePlugin.DEBUG_GRAPH_EXTRACTOR,
								refTypeName + " already added!");

					}
				}
				
			} catch (ObjectCollectedException doNothing) {
				// If the type is already collected by the discard it				
			}

		}

		return (IJavaReferenceType[]) result
				.toArray(new IJavaReferenceType[result.size()]);
	}

	private static JDIDebugTarget getDebugTarget(IJavaThread thread)
			throws MemoryAccessException {
		JDIDebugTarget result = null;

		// Attention: use the inner class because the interface does not
		// support the access to the virtual machine
		if (!(thread.getDebugTarget() instanceof JDIDebugTarget)) {
			throw new MemoryAccessException("Cannot access the classes, "
					+ "because the debug target of the thread "
					+ "is not an instance of JDIDebugTarget.");
		}

		result = (JDIDebugTarget) thread.getDebugTarget();

		return result;
	}

	private static IJavaReferenceType createNewReferenceType(
			ReferenceType refType, JDIDebugTarget debugTarget) {
		assert refType != null;
		assert debugTarget != null;

		IJavaReferenceType result = null;

		if (refType instanceof ArrayType) {
			result = new JDIArrayType(debugTarget, (ArrayType) refType);
		} else if (refType instanceof ClassType) {
			result = new JDIClassType(debugTarget, (ClassType) refType);
		} else if (refType instanceof InterfaceType) {
			result = new JDIInterfaceType(debugTarget, (InterfaceType) refType);
		} else {
			logger.log(IStatus.ERROR, "Cannot convert class "
					+ refType.getClass().getName()
					+ " in one of the following classes "
					+ ArrayType.class.getName() + ", "
					+ ClassType.class.getName() + " and "
					+ InterfaceType.class.getName());
		}

		return result;
	}

	public static IJavaVariable getVariable(IJavaThread thread, Path path,
			int classLoaderID) throws MemoryAccessException {
		assert thread != null;
		assert thread.isSuspended();

		IJavaVariable result = null;

		int pathSize = path.size();
		int pathIndex = 0;

		IGraphEdge pathEdge = path.get(pathIndex);

		if (pathEdge instanceof ITypeEdge) {
			// The path must contains at least two element
			// * The first entry has to point to the type (e.g.
			// java.lang.String)
			// * The second entry has to point to a static attribute of the type
			//
			// Additional entries are possible if the second entry is an object
			// reference.
			assert pathSize >= 2;

			IJavaReferenceType refType = MemoryAccessUtil
					.getTypeForRefTypeEdge((ITypeEdge) pathEdge,
							thread, classLoaderID);

			pathEdge = path.get(1);
			// the second edge must instance of IGlobalVariableEdge
			assert pathEdge instanceof IGlobalVariableEdge;

			result = MemoryAccessUtil.getVariableForGlobalVarEdge(
					(IGlobalVariableEdge) pathEdge, refType);

			pathIndex = 2;

		} else if (pathEdge instanceof ILocalVariableEdge) {
			result = MemoryAccessUtil.getVariableForLocalVarEdge(
					(ILocalVariableEdge) pathEdge, thread);

			pathIndex = 1;
		}

		for (; pathIndex < pathSize; pathIndex++) {
			IJavaValue value = null;
			pathEdge = path.get(pathIndex);

			try {
				value = (IJavaValue) result.getValue();
			} catch (DebugException e) {
				throw new MemoryAccessException(
						"Failure on acces value of the current variable", e);
			}

			if (pathEdge instanceof IArrayElementEdge
					&& value instanceof IJavaArray) {
				result = MemoryAccessUtil.getVariableForArrayElementEdge(
						(IArrayElementEdge) pathEdge, (IJavaArray) value);

			} else if (pathEdge instanceof IObjectFieldEdge
					&& value instanceof IJavaObject) {
				result = MemoryAccessUtil.getVariableForObjRefFieldEdge(
						(IObjectFieldEdge) pathEdge,
						(IJavaObject) value);

			}

		}

		if (result == null) {
			throw new MemoryAccessException(
					"Cannot retrieve variable for path:\n" + path);
		}

		return result;
	}

	/**
	 * Returns for the given edge and the give object the field defined by the
	 * edge.
	 * 
	 * @param edge
	 *            the edge including the field name and declaring type name of
	 *            the field.
	 * @param object
	 *            the value of the object.
	 * @return for the given edge and the give object the field defined by the
	 *         edge.
	 * @throws MemoryAccessException
	 *             the method fails. Reasons include:
	 *             <ul>
	 *             <li>The extraction of the defined field failed.</li>
	 *             </ul>
	 */
	public static IJavaVariable getVariableForObjRefFieldEdge(
			IObjectFieldEdge edge, IJavaObject object)
			throws MemoryAccessException {
		IJavaVariable result = null;

		String fieldName = edge.getFieldName();
		String declTypeSignature = edge.getDeclaringTypeSignature();

		try {
			result = object.getField(fieldName, declTypeSignature);
		} catch (DebugException e) {
			throw new MemoryAccessException("Failure on extracting field '"
					+ fieldName + "' of the declarint type '"
					+ edge.getDeclaringTypeName() + "'", e);
		}

		return result;
	}

	/**
	 * Returns for the given edge and the given array the array element defined
	 * by the edge.
	 * 
	 * @param edge
	 *            the edge include the array index of the array element we are
	 *            looking for.
	 * @param array
	 *            the value of the array.
	 * @return for the given edge and the given array the array element defined
	 *         by the edge.
	 * @throws MemoryAccessException
	 *             the method fails. Reasons include:
	 *             <ul>
	 *             <li>The access of the array length failed.</li>
	 *             <li>The access of all array variables failed.</li>
	 *             <li>The variables found for the variable name not exists or
	 *             is amibigous.</li>
	 *             </ul>
	 */
	public static IJavaVariable getVariableForArrayElementEdge(
			IArrayElementEdge edge, IJavaArray array)
			throws MemoryAccessException {
		IJavaVariable result = null;

		int arrayIndex = edge.getArrayIndex();
		int targetArrayLength = -1;

		try {
			targetArrayLength = array.getLength();
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Cannot retrieve the length of the given array.", e);
		}

		if (arrayIndex < 0 || arrayIndex >= targetArrayLength) {
			throw new MemoryAccessException("The defined array index "
					+ arrayIndex + " is out of range.");
		}

		IVariable[] variables = null;

		try {
			variables = array.getVariables();
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Access of all variables in the array failed", e);
		}

		IVariable[] variablesByName = MemoryAccessUtil.searchVariableByName(
				variables, "[" + arrayIndex + "]");

		if (variablesByName.length != 1) {
			throw new MemoryAccessException("Variable wiht name '["
					+ arrayIndex + "]' not found or not unique");
		}

		result = (IJavaVariable) variablesByName[0];

		return result;
	}

	/**
	 * Returns for the given edge and reference type the variable defined by the
	 * edge.
	 * 
	 * @param edge
	 *            the edge include the field or variable name that we are
	 *            looking for.
	 * @param refType
	 *            the reference type.
	 * @return for the given edge and reference type the variable defined by the
	 *         edge.
	 * @throws MemoryAccessException
	 *             MemoryAccessException the method fails. Reasons include:
	 *             <ul>
	 *             <li>The extraction of the field by the field name failed.</li>
	 *             <li>The field was not found in the given reference type.</li>
	 *             </ul>
	 */
	public static IJavaVariable getVariableForGlobalVarEdge(
			IGlobalVariableEdge edge, IJavaReferenceType refType)
			throws MemoryAccessException {
		IJavaVariable result = null;

		String fieldName = edge.getName();

		try {
			result = refType.getField(fieldName);
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Cannot extract the field with name " + fieldName, e);
		}
		if (result == null) {
			throw new MemoryAccessException("Cannot find the field with name "
					+ fieldName);
		}

		return result;
	}

	/**
	 * Returns for the given edge the reference type in the thread.
	 * 
	 * @param edge
	 *            the reference type edge include all informations about the
	 *            referenced type.
	 * @param thread
	 *            a suspended thread in the java virtual machine.
	 * @param classLoaderID
	 *            the unique id of the class loader
	 * @return for the given edge the reference type in the thread.
	 * @throws MemoryAccessException
	 *             the method fails. Reasons include:
	 *             <ul>
	 *             <li>The search of the reference type by his name failed.</li>
	 *             </ul>
	 */
	public static IJavaReferenceType getTypeForRefTypeEdge(
			ITypeEdge edge, IJavaThread thread, int classLoaderID)
			throws MemoryAccessException {
		IJavaReferenceType result = null;

		String refTypeName = edge.getTypeName();

		result = MemoryAccessUtil.searchReferenceTypeByName(refTypeName,
				classLoaderID, thread);

		return result;
	}

	/**
	 * Returns for the given edge the described java variable.
	 * 
	 * @param edge
	 *            the local variable edge include all informations about the
	 *            stack frame and the variable name.
	 * @param thread
	 *            a suspended thread in the java virtual machine.
	 * @return for the given edge the described java variable.
	 * @throws MemoryAccessException
	 *             the method fails. Reasons include:
	 *             <ul>
	 *             <li>The stack frame number not exists in the given thread.</li>
	 *             <li>The extracted frame is not an instance of
	 *             IJavaStackFrame.</li>
	 *             <li>The extraction of all variables in the stack frame
	 *             failed.</li>
	 *             <li>The variables found for the variable name not exists or
	 *             is amibigous.</li>
	 *             </ul>
	 */
	public static IJavaVariable getVariableForLocalVarEdge(
			ILocalVariableEdge edge, IJavaThread thread)
			throws MemoryAccessException {
		IJavaVariable result = null;

		int frameNumber = edge.getStackFrame().getFrameNumber();
		String varName = edge.getVariableName();

		IStackFrame frame = null;
		try {
			frame = thread.getStackFrames()[frameNumber];
		} catch (DebugException e) {
			throw new MemoryAccessException("Cannot receive frame nr. "
					+ frameNumber + " of the given thread", e);
		}

		if (!(frame instanceof IJavaStackFrame)) {
			throw new MemoryAccessException("Stackframe " + frameNumber
					+ " is not an instance of IJavaStackFrame");
		}

		IVariable[] variables = null;
		try {
			variables = frame.getVariables();
		} catch (DebugException e) {
			throw new MemoryAccessException(
					"Cannot extract the variables of the frame", e);
		}

		IVariable[] variablesByName = MemoryAccessUtil.searchVariableByName(
				variables, varName);

		if (variablesByName.length != 1) {
			throw new MemoryAccessException("Variable wiht name '" + varName
					+ "' not found or not unique");
		}

		result = (IJavaVariable) variablesByName[0];

		return result;
	}

	public static IJavaReferenceType searchReferenceTypeByName(String typeName,
			int classLoaderID, IJavaThread thread) throws MemoryAccessException {
		IJavaReferenceType result = null;

		List classes = MemoryAccessUtil
				.getLoadedClassesByName(thread, typeName);

		ReferenceType refType = null;

		if (classes.size() == 1) {
			// if the class is only loaded one time
			refType = (ReferenceType) classes.get(0);

		} else {
			// , else look for the class with the corresponding class loader ID
			for (Iterator itClass = classes.iterator(); itClass.hasNext();) {
				refType = (ReferenceType) itClass.next();

				int refTypeClassLoaderID = MemoryAccessUtil
						.getUniqueClassLoaderID(refType);

				if (classLoaderID == refTypeClassLoaderID) {
					break;
				}
			}
		}

		if (refType != null) {
			JDIDebugTarget debugTarget = MemoryAccessUtil
					.getDebugTarget(thread);
			result = MemoryAccessUtil.createNewReferenceType(refType,
					debugTarget);

		}

		if (result == null) {
			throw new MemoryAccessException(
					"Reference type for the given parameters not found");
		}

		return result;
	}

	private final static int getUniqueClassLoaderID(ReferenceType classRef) {
		int result = -1;

		ReferenceType clRefType = classRef.classLoader().referenceType();

		try {
			// FIXME introduce a smarter solution
			Field idField = clRefType.fieldByName("currentClassLoaderID");
			Value idValue = classRef.classLoader().getValue(idField);
			result = new Integer(idValue.toString()).intValue();
		} catch (Throwable doNothing) {
			// do nothing
		}

		return result;
	}

	private static IVariable[] searchVariableByName(IVariable[] variables,
			String varName) throws MemoryAccessException {
		IVariable[] result = null;

		Collection colResult = new Vector();

		String intVarName = null;

		for (int i = 0; i < variables.length; i++) {
			IVariable variable = variables[i];

			try {
				intVarName = variable.getName();
			} catch (DebugException e) {
				throw new MemoryAccessException(
						"Cannot read the name of the variable", e);
			}

			if (intVarName.equals(varName)) {
				colResult.add(variable);
			}
		}

		result = (IVariable[]) colResult
				.toArray(new IVariable[colResult.size()]);

		return result;
	}

}