/*
 * $Id: DDLogger.java,v 1.8 2005/08/23 11:21:39 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 03.01.2005 18:04:09 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.logger;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;

/**
 * COMMENTME Comment type DDLogger
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DDLogger {

	private Plugin fPlugin = null;

	private Class fTracer = null;

	// this variables will be used to measure und userfriendly print the
	// duration of processes
	private final static Map fTimeMap = new HashMap();

	private final static SimpleDateFormat DATE_FORMAT_TRACE = new SimpleDateFormat(
			"HH:mm:ss:SSS");

	/**
	 * 
	 * @param plugin
	 * @param tracer
	 *            the tracer is the class that have to log or trace messages.
	 */
	public DDLogger(Plugin plugin, Class tracer) {
		this.fPlugin = plugin;
		this.fTracer = tracer;
	}

	/**
	 * Log the message with the specified severity (IStatus.INFO,
	 * IStatus.WARNUNG, IStatus.ERROR)
	 * 
	 * @param severity
	 *            the severity of the log message.
	 * @param message
	 *            the message you want to log.
	 * @return the status of the log message or null if the plugin is not
	 *         available.
	 */
	public IStatus log(int severity, String message) {
		return this.log(severity, message, null);
	}

	/**
	 * Log the message with the given exception and the specified severity
	 * (IStatus.INFO, IStatus.WARNING, IStatus.ERROR). As status code thif
	 * method logs <code>IStatus.OK</code>.
	 * 
	 * @param severity
	 *            the severity of the log message.
	 * @param message
	 *            the message you want to log.
	 * @param exception
	 *            the additional exception to the log message.
	 * @return the status of the log message or null if the plugin is not
	 *         available.
	 */
	public IStatus log(int severity, String message, Throwable exception) {
		return this.log(severity, IStatus.OK, message, exception);
	}

	/**
	 * Log the message with the given exception and the specified severity
	 * (IStatus.INFO, IStatus.WARNING, IStatus.ERROR).
	 * 
	 * @param severity
	 *            the severity of the log message.
	 * @param statuCode
	 *            the plugin or module specific status code.
	 * @param message
	 *            the message you want to log.
	 * @param exception
	 *            the additional exception to the log message.
	 * @return the status of the log message or null if the plugin is not
	 *         available.
	 */
	public IStatus log(int severity, int statusCode, String message,
			Throwable exception) {
		IStatus result = null;

		result = new Status(severity, this.fPlugin.getBundle()
				.getSymbolicName(), statusCode, message, exception);

		this.log(result);

		return result;
	}

	/**
	 * Log the status.
	 * 
	 * @param status
	 *            the status to log
	 */
	public void log(IStatus status) {
		if (this.fPlugin != null) {
			this.fPlugin.getLog().log(status);

		} else {
			int severity = status.getSeverity();
			String message = status.getMessage();
			Throwable exception = status.getException();

			if (severity == IStatus.INFO) {
				System.out.println(message);
			} else {
				System.err.println(message);
			}

			if (exception != null) {
				exception.printStackTrace();
			}
		}
	}

	/**
	 * Trace the given message to the standard output if tracing is enabled.
	 * 
	 * @param debug
	 *            use predefined public static variables to indicate if you want
	 *            to trace the message or not.
	 * @param tracer
	 *            the tracer is the class that trace the given message.
	 * @param message
	 *            the message to trace.
	 */
	public void trace(boolean debug, String message) {
		if (this.fPlugin.isDebugging() && debug) {
			Logger.getLogger(this.fTracer.getName()).debug(message);
		}
	}

	/**
	 * Trace the given message and exception to the standard output if tracing
	 * is enabled.
	 * 
	 * @param debug
	 *            use predefined public static variables to indicate if you want
	 *            to trace the message or not.
	 * @param tracer
	 *            the tracer is the class that trace the given message.
	 * @param message
	 *            the message to trace.
	 * @param exception
	 *            the exception to trace.
	 */
	public void trace(boolean debug, String message, Throwable exception) {
		if (this.fPlugin.isDebugging() && debug) {
			Logger.getLogger(this.fTracer.getName()).debug(message, exception);
		}
	}

	/**
	 * Trace the message and print the actual time as prefix of the message.
	 * 
	 * @param debug
	 *            use predefined public static variables to indicate if you want
	 *            to trace the message or not.
	 * @param message
	 *            the message to trace.
	 * @return the difference, measured in milliseconds, between the current
	 *         time and midnight, January 1, 1970 UTC.
	 */
	public long traceAndTime(boolean debug, String message) {
		long time = System.currentTimeMillis();

		String strTime = DDLogger.DATE_FORMAT_TRACE.format(new Date(time));

		this.trace(debug, strTime + " - " + message);

		return time;
	}

	/**
	 * Trace the message with the current time and use the key to stop the stop
	 * watch. After that prints the duration meassured by the stop watch. It is
	 * recommanded that you have start the stop watch with
	 * <code>traceAndTimeStart</code> with the same key.
	 * 
	 * @param debug
	 *            use predefined public static variables to indicate if you want
	 *            to trace the message or not.
	 * @param key
	 *            unique identifier for the stop watch.
	 * @param message
	 *            the message to trace.
	 * @return the difference, measured in milliseconds, between the current
	 *         time and midnight, January 1, 1970 UTC.
	 */
	public long traceAndTimeEnd(boolean debug, String key, String message) {
		long endTime = this.traceAndTime(debug, message);
		long startTime = -1;
		long timeDiff = -1;

		String uKey = this.getUniqueKey(key);

		if (DDLogger.fTimeMap.containsKey(uKey)) {
			startTime = ((Long) DDLogger.fTimeMap.get(uKey)).longValue();
			timeDiff = endTime - startTime;

			long ms = timeDiff % 1000;
			long sec = (timeDiff / 1000) % 60;
			long min = timeDiff / 60000;

			this.trace(debug, "Duration " + timeDiff + " ms ("
					+ this.paddingString("" + min, 2, '0') + ":"
					+ this.paddingString("" + sec, 2, '0') + "."
					+ this.paddingString("" + ms, 3, '0') + ")");
		}

		return endTime;
	}

	/**
	 * Fill the given string with the given filling symbol until the required
	 * size is reached. The filling symbol will append in front of the given
	 * string.
	 * 
	 * @param toFill
	 * @param requiredSize
	 * @param fillingSymbol
	 * @return the with filling symbol extended string
	 */
	private String paddingString(String toFill, int requiredSize,
			char fillingSymbol) {
		if (toFill.length() < requiredSize) {
			toFill = fillingSymbol + toFill;
			return this.paddingString(toFill, requiredSize, fillingSymbol);
		}

		return toFill;
	}

	/**
	 * Trace the message with the current time and use the key to start the stop
	 * watch. Use <code>traceAndTimeEnd</code> with the same key to stop the
	 * stop watch and print the duration of the two calls.
	 * 
	 * @param debug
	 *            use predefined public static variables to indicate if you want
	 *            to trace the message or not.
	 * @param key
	 *            unique identifier for the stop watch.
	 * @param message
	 *            the message to trace.
	 * @return the difference, measured in milliseconds, between the current
	 *         time and midnight, January 1, 1970 UTC.
	 */
	public long traceAndTimeStart(boolean debug, String key, String message) {
		long startTime = this.traceAndTime(debug, message);
		String uKey = this.getUniqueKey(key);

		DDLogger.fTimeMap.put(uKey, new Long(startTime));

		return startTime;
	}

	/**
	 * Generate an unique key including the plugin name, the tracer name and the
	 * given key.
	 * 
	 * @param key
	 * @return unique key including plugin name, tracer name and the given key.
	 */
	private String getUniqueKey(String key) {
		String result = this.fPlugin.getClass().getName() + "-"
				+ this.fTracer.getName() + "-" + key;

		return result;
	}

}