/*
 * $Id: GraphEdge.java,v 1.35 2005/11/29 16:05:57 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 26.08.2004 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.GraphEdgeComparator;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.memorygraph.delta.DeltaTreeFactory;
import org.deltadebugging.ddstate.memorygraph.delta.delete.DeleteObjectReferenceDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateObjectReferenceDelta;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.deltadebugging.ddstate.memorygraph.value.INullValueNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;

/**
 * This class represented a directed graph edge. The name of the graph edge
 * includes the operation by which you can retrieve the target node starting
 * from source node.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class GraphEdge extends GraphElement implements IGraphEdge {

	/**
	 * The start point of the edge.
	 */
	private IGraphNode fSourceNode;

	/**
	 * The end point of the edge.
	 */
	private IGraphNode fTargetNode;

	/**
	 * This name will given to the graph element if the name was unspecified
	 */
	private static final String ELEMENT_NAME_UNDEFINED = "<ELEMENT UNDEFINED>";

	/**
	 * A name of the graphnode. The default is <code>NODE_NAME_UNDEFINED</code>
	 * but have to be overriden at construction.
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.GraphEdge#ELEMENT_NAME_UNDEFINED
	 */
	public String fName = ELEMENT_NAME_UNDEFINED;

	/**
	 * Create a new directed graph edge.
	 * 
	 * @param rootNode
	 *            the root node of this graph element.
	 * @param sourceNode
	 *            the source node where the edge points from.
	 * @param targetNode
	 *            the target node where the edge points to.
	 * @param name
	 *            the name of the edge. The name is the name of the operation by
	 *            which you can obtain the target node from the source node.
	 */
	public GraphEdge(IRootNode rootNode, IGraphNode sourceNode,
			IGraphNode targetNode, String name) {
		super(rootNode);
		this.fTargetNode = targetNode;
		this.fSourceNode = sourceNode;
		this.fName = name;

		if (DDStatePlugin.DEBUG_GRAPH_STATISTICS && rootNode != null) {
			rootNode.getStatistics().numberOfGraphEdges++;
		}
	}
    
    /**
     * Create a new directed graph edge.
     * 
     * @param sourceNode
     *            the source node where the edge points from.
     * @param targetNode
     *            the target node where the edge points to.
     * @param name
     *            the name of the edge. The name is the name of the operation by
     *            which you can obtain the target node from the source node.
     */
    public GraphEdge(IGraphNode sourceNode,
            IGraphNode targetNode, String name) {
        this(sourceNode.getRootNode(), sourceNode, targetNode, name);
    }

	/**
	 * Returns the target node of the edge.
	 * 
	 * @return the target node of the edge.
	 */
	public IGraphNode getTargetNode() {
		return this.fTargetNode;
	}

	/**
	 * Returns the source node of the edge.
	 * 
	 * @return the source node of the edge.
	 */
	public IGraphNode getSourceNode() {
		return this.fSourceNode;
	}

	/**
	 * @return the name of the graph element if it was specified. Else the
	 *         result equals <code>ELEMENT_NAME_UNDEFINED</code>.
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.GraphElement#ELEMENT_NAME_UNDEFINED
	 */
	public String getName() {
		return this.fName;
	}

	public boolean graphEquals(IGraphElement element) {
		if (!(element instanceof IGraphEdge)) {
			return false;
		}

		IGraphEdge objEdge = (IGraphEdge) element;
		IGraphNode objSourceNode = objEdge.getSourceNode();

		boolean result = false;

		result = this.getName().equals(objEdge.getName())
				&& this.getSourceNode().graphEquals(objSourceNode);

		return result;
	}

	public String toString() {
		String result = "GraphEdge: \n" + "\tSourceNode: "
				+ this.fSourceNode.toString() + "\n" + "\tOperation: "
				+ this.getName() + "\n" + "\tTargetNode: "
				+ this.getTargetNode().toString() + "\n";
		return result;
	}

	public DeltaSet computeDeltaSet(IGraphElement element)
			throws DeltaComputationException {

		if (!this.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
			return new DeltaSet();
		}

		DeltaSet result = new DeltaSet();

		assert element instanceof IGraphEdge;
		IGraphEdge edge = (IGraphEdge) element;

		assert this.getSourceNode().isAssociatedWith(edge.getSourceNode());
		assert GraphEdgeComparator.getInstance().compare(this, edge) == 0;

        // use flag to find scenarios which are not handled
        boolean computationValid = false;
        
		if (this.getTargetNode().isAssociatedWith(edge.getTargetNode())) {
            computationValid = true;
            
			// mask as equal
			GraphDecorator.decorateGraphElementsAsEqual(this, element);

			result.addAll(this.getTargetNode().computeDeltaSet(
					edge.getTargetNode()));

		} else {
      
            if (edge.getTargetNode() instanceof INullValueNode) {                
                assert this.getTargetNode() instanceof IObjectNode;                
                computationValid = true;                
                
                // mask nodes
                GraphDecorator.decorateGraphElementsAsDifferent(this, edge);
                GraphDecorator.decorateGraphNodeAs(edge.getTargetNode(),
                        GraphDecorator.ONLY_IN_SECONDGRAPH);

                result.add(new DeleteObjectReferenceDelta(this));                
            
            } else if (edge.getTargetNode() instanceof IObjectNode) {

                // mask as changed
                GraphDecorator.decorateGraphElementsAsDifferent(this, edge);
                
                if (edge.getTargetNode().hasAssociatedElement()) {
                    assert !edge.getTargetNode().isAssociatedWith(this.getTargetNode());
                    computationValid = true;

                    // reference a node of the common subgraph
                    result.add(new UpdateObjectReferenceDelta(this,
                            (IObjectNode) edge.getTargetNode()
                                    .getAssociatedElement()));
                } else {
                    computationValid = true;
                    
                    // reference a new node
                    result.addAll(DeltaTreeFactory.createDeltaTree(this,
                            (IObjectNode) edge.getTargetNode()));
                }
                
            }
            
            
        }
            
		if (!computationValid) {
			throw new DeltaComputationException(
					"This should never be called. Failure in program. Inform me (KarstenLehmann@web.de)");
		}

		return result;
	}
}