/*
 * $Id: GraphElement.java,v 1.16 2005/11/24 08:31:18 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 01.09.2004 13:35:42 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.eclipse.core.runtime.IStatus;

/**
 * COMMENTME Comment type GraphElement
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class GraphElement implements IGraphElement {

	// Personal logger for this class
	private final static DDLogger logger = DDStatePlugin
			.getLogger(GraphElement.class);

	/**
	 * This maks is used by the graph comparison. We remember if this graph
	 * element is already compared, exists only in the first or second graph,
	 * exists in both graph and is equal or differnt. The default value is
	 * <code>GraphDecorator.NOT_COMPARED</code>.
	 * 
	 * @see GraphDecorator
	 */
	private byte fDifferenceMask = GraphDecorator.NOT_COMPARED;

	/**
	 * Every graph element knows about his rootnode. The rootnode support
	 * additional informations for all graphelements. This informations p.e.
	 * will be used by creating (unfolding) the graph to avoid creating two
	 * nodes of the same object.
	 */
	IRootNode fRootNode = null;

	private IGraphElement fAssociatedElement = null;

	private boolean fDiffVisualizerFlag;

	public GraphElement(IRootNode rootNode) {
		this.fRootNode = rootNode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#getRootNode()
	 */
	public IRootNode getRootNode() {
		return this.fRootNode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.mcs.IMCSComparable#resetAssociatedElement()
	 */
	public void resetAssociatedElement() {
		this.fAssociatedElement = null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#setAssociatedElement(org.deltadebugging.ddstate.memorygraph.IGraphElement)
	 */
	public void setAssociatedElement(IGraphElement graphElement) {
		if (graphElement == null)
			throw new IllegalArgumentException(
					"Given graph element must not be null.");

		if (!graphElement.getClass().getName()
				.equals(this.getClass().getName())) {

			logger.log(IStatus.WARNING, "Given graph element ("
					+ graphElement.getClass().getName()
					+ ") is not of the same type as this graph element ("
					+ this.getClass().getName() + ")");

			throw new IllegalArgumentException(
					"Given graph element is not of the same type as this graph element");

		}

		this.fAssociatedElement = graphElement;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.IGraphElementComparable#setDiffVisualizerFlag()
	 */
	public void setDiffVisualizerFlag() {
		this.fDiffVisualizerFlag = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.IGraphElementComparable#isDiffVisualizerFlagSet()
	 */
	public boolean isDiffVisualizerFlagSet() {
		return this.fDiffVisualizerFlag == true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#getAssociatedElement()
	 */
	public IGraphElement getAssociatedElement() {
		return this.fAssociatedElement;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#isAssociatedWith(IGraphElement)
	 */
	public boolean isAssociatedWith(IGraphElement graphElement) {
		boolean result = true;

		result &= this.hasAssociatedElement();
		result &= graphElement.hasAssociatedElement();
		result &= this.getAssociatedElement() == graphElement;
		result &= graphElement.getAssociatedElement() == this;

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#hasAssociatedElement()
	 */
	public boolean hasAssociatedElement() {
		return this.fAssociatedElement != null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.delta.IDeltaComparable#maskAs(int)
	 */
	public void maskAs(byte mask) {
		this.fDifferenceMask = mask;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.delta.IDeltaComparable#isMaskedAs(int)
	 */
	public boolean isMaskedAs(byte mask) {
		return ((this.fDifferenceMask & mask) > 0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.comparator.delta.IDeltaComparable#getDifferenceMask()
	 */
	public int getDifferenceMask() {
		return this.fDifferenceMask;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#computeDeltaSet(org.deltadebugging.ddstate.memorygraph.IGraphElement)
	 */
	public DeltaSet computeDeltaSet(IGraphElement element)
			throws DeltaComputationException {
		throw new DeltaComputationException("Not implemented for "
				+ this.getClass().getName());
	}
}