/*
 * $Id: DeltaTreeFactory.java,v 1.5 2005/11/28 11:12:49 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 05.05.2005 16:47:58 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.delta;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphNode;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreateObjectDelta;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreatePrimitiveValueDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateObjectReferenceDelta;
import org.deltadebugging.ddstate.memorygraph.value.INullValueNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IPrimitiveValueNode;
import org.deltadebugging.ddstate.memorygraph.value.IValueNode;

/**
 * The delta tree factory is a usefull util to generate deltas for subtrees
 * existing only in one memory graph. With this util we can create all necessary
 * deltas to build the existing subtree on the one memory graph in the second
 * one.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DeltaTreeFactory {

	private DeltaTreeFactory() {
		// use the static functions that provide this factory
	}

	/**
	 * Create a delta set of all deltas to transfer the subtree in the first
	 * memory graph to the second one. The given edge has to exists in both
	 * graphs. The given graph node must only exists in the first graph must not
	 * have an associated element.
	 * 
	 * @param edgeToChange
	 * @param valueToCreate
	 * @return a delta set of all deltas that are necessary to generate the
	 *         values and adapting the references of the second memory graph so
	 *         it look like the first memory graph for the subtree of the given
	 *         graph node.
	 */
	public static DeltaSet createDeltaTree(IGraphEdge edgeToChange,
			IObjectNode valueToCreate) {
		DeltaSet result = new DeltaSet();

		assert !(valueToCreate instanceof INullValueNode);
		assert !valueToCreate.hasAssociatedElement();

		IObjectNode objToCreate = valueToCreate;
		CreateObjectDelta createDelta = null;

		if (!existsDeltaFor(objToCreate)) {
			createDelta = new CreateObjectDelta(objToCreate);
			registerDelta(objToCreate, createDelta);

			result.add(createDelta);
			result.addAll(computeDependendDeltas(objToCreate));

		} else {
			createDelta = getDeltaFor(valueToCreate);

		}

		result.add(new UpdateObjectReferenceDelta(edgeToChange, createDelta));

		assert !result.isEmpty();

		return result;
	}

	private static DeltaSet generateCreateDelta(CreateObjectDelta sourceDelta,
			IGraphEdge edgeToChange, IObjectNode valueToCreate) {
		DeltaSet result = new DeltaSet();

		assert !(valueToCreate instanceof INullValueNode);
		assert !valueToCreate.hasAssociatedElement();

		CreateObjectDelta createDelta = null;

		if (!existsDeltaFor(valueToCreate)) {
			createDelta = new CreateObjectDelta(valueToCreate);
			registerDelta(valueToCreate, createDelta);

			result.add(createDelta);
			result.addAll(computeDependendDeltas(valueToCreate));

		} else {
			createDelta = getDeltaFor(valueToCreate);

		}

		result.add(new UpdateObjectReferenceDelta(sourceDelta, edgeToChange,
				createDelta));

		return result;
	}

	/**
	 * Compute the dependend deltas for all the successors of the given object
	 * reference. Returns a delta set containing all computed deltas. <br>
	 * <b>Note: </b> Precondition is that for the given object reference already
	 * a create delta was computed and registered.
	 * 
	 * @param objRef
	 * @return a delta set containing all computed deltas.
	 */
	private static DeltaSet computeDependendDeltas(
			IObjectNode objRef) {
		DeltaSet result = new DeltaSet();

		CreateObjectDelta createDelta = null;
		createDelta = getDeltaFor(objRef);
		assert createDelta != null;

		Iterator itSuccessors = objRef.getChildren().iterator();

		while (itSuccessors.hasNext()) {
			IGraphEdge graphEdge = (IGraphEdge) itSuccessors.next();
			result.addAll(computeDependendDeltas(createDelta, graphEdge));
		}

		return result;
	}

	/**
	 * Compute all dependent deltas for the given create delta and the graph
	 * edge.
	 * 
	 * @param createDelta
	 * @param newEdge
	 * @return
	 */
	private static DeltaSet computeDependendDeltas(
			CreateObjectDelta createDelta, IGraphEdge newEdge) {

		DeltaSet result = new DeltaSet();

		IGraphNode newTargetNode = newEdge.getTargetNode();

		if (newTargetNode instanceof INullValueNode) {
			// do nothing

		} else if (newTargetNode instanceof IPrimitiveValueNode) {
			CreatePrimitiveValueDelta newDelta = new CreatePrimitiveValueDelta(
					createDelta, newEdge,
					(IPrimitiveValueNode) newTargetNode);

			result.add(newDelta);

		} else if (newTargetNode instanceof IObjectNode) {

			if (newTargetNode.hasAssociatedElement()) {
				UpdateObjectReferenceDelta newDelta = new UpdateObjectReferenceDelta(
						createDelta, newEdge,
						(IObjectNode) newTargetNode
								.getAssociatedElement());

				result.add(newDelta);

			} else if (newTargetNode instanceof IObjectNode) {
				result.addAll(generateCreateDelta(createDelta, newEdge,
						(IObjectNode) newTargetNode));

			}
		}

		return result;
	}

	// **********************************************************************************
	/** Use the map to register, all the creation deltas */
	private static Map fDDRunCreationMaps = new HashMap();

	/**
	 * Register that the given graph node was created and remember the delta for
	 * that.
	 * 
	 * @param graphNode
	 *            the graph node that will be created
	 * @param delta
	 *            the delta to create the graph node
	 */
	private static void registerDelta(IValueNode graphNode,
			CreateObjectDelta delta) {
		int ddRunID = graphNode.getRootNode().getDDRunID();
		Map ddRunCreationMap = getMapForDDRunID(ddRunID);

		ddRunCreationMap.put(graphNode, delta);
	}

	/**
	 * Returns the map for the given dd run id. If not a map is registered for
	 * the given dd run id a new map will created and registerd.
	 * 
	 * @param ddRunID
	 *            the id of the delta debugging run.
	 * @return the existing map for the given dd run id or a new map if no map
	 *         exists so far.
	 */
	private static Map getMapForDDRunID(int ddRunID) {
		Map result = (Map) fDDRunCreationMaps.get(new Integer(ddRunID));

		if (result == null) {
			result = new HashMap();
			fDDRunCreationMaps.put(new Integer(ddRunID), result);
		}

		return result;
	}

	/**
	 * Look up if for the given graph node a delta is already created. If it is
	 * so return true, otherwise false.
	 * 
	 * @param graphNode
	 *            the graph node for look up.
	 * @return true if for the given graph node a delta is already created,
	 *         otherwise false.
	 */
	private static boolean existsDeltaFor(IValueNode graphNode) {
		int ddRunID = graphNode.getRootNode().getDDRunID();
		Map ddRunCreationMap = getMapForDDRunID(ddRunID);

		return ddRunCreationMap.containsKey(graphNode);
	}

	/**
	 * Returns the delta for to create the given graph node if one exists,
	 * otherwise return <code>null</code>.
	 * 
	 * @param graphNode
	 *            the graph node for look up.
	 * @return the delta for to create the given graph node if one exists,
	 *         otherwise return <code>null</code>.
	 */
	private static CreateObjectDelta getDeltaFor(IValueNode graphNode) {
		int ddRunID = graphNode.getRootNode().getDDRunID();
		Map ddRunCreationMap = getMapForDDRunID(ddRunID);

		return (CreateObjectDelta) ddRunCreationMap.get(graphNode);
	}

	/**
	 * Clear the internal used cache for the given delta debugging run id.
	 * 
	 * @param ddRunID
	 *            the deltat debugging run id.
	 */
	public static void clearCache(int ddRunID) {
		fDDRunCreationMaps.put(new Integer(ddRunID), null);
	}
}
