/*
 * $Id: MemoryGraphDelta.java,v 1.6 2005/11/21 10:42:04 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 29.07.2005 18:14:25 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.delta;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.memorygraph.delta.location.IDeltaLocation;

/**
 * The memory graph delta describes differences between two memory graphs. A
 * delta also supports methods to transfer one memory graph into another.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class MemoryGraphDelta implements IMemoryGraphDelta {

	/** A collection of deltas that are necessary that this delta can applied */
	private final Collection fDependencies = new Vector();

	/**
	 * The location where this delta should be applied. E.g this is a path to a
	 * well defined position in the memory graph or another delta (e.g. a
	 * CreateObjectDelta).
	 */
	private IDeltaLocation fLocation = null;

	public MemoryGraphDelta(IDeltaLocation location) {
		this.fLocation = location;
	}

	public MemoryGraphDelta() {
		this.fLocation = null;
	}

	protected void setDeltaLocation(IDeltaLocation location) {
		this.fLocation = location;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#addDependeny(org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta)
	 */
	public void addDependeny(IMemoryGraphDelta delta) {
		if (!this.fDependencies.contains(delta) && delta != null) {
			this.fDependencies.add(delta);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#containsAllDependencies(org.deltadebugging.ddcore.DeltaSet)
	 */
	public boolean containsAllDependencies(DeltaSet deltaSet) {
		boolean result = true;

		Iterator itDeltas = this.fDependencies.iterator();

		while (itDeltas.hasNext() && result == true) {
			IMemoryGraphDelta delta = (IMemoryGraphDelta) itDeltas.next();

			if (!deltaSet.contains(delta)) {
				result = false;
			} else {
				result &= delta.containsAllDependencies(deltaSet);
			}
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#getAllDependentDeltas()
	 */
	public Collection getAllDependentDeltas() {
		Collection result = this.fDependencies;

		for (Iterator itDependencies = result.iterator(); itDependencies
				.hasNext();) {
			IMemoryGraphDelta delta = (IMemoryGraphDelta) itDependencies.next();
			result.addAll(delta.getAllDependentDeltas());
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#getLocation()
	 */
	public IDeltaLocation getLocation() {
		if (this.fLocation == null) {
			throw new IllegalStateException("Location not set.");
		}

		return this.fLocation;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#isDependentOn(org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta)
	 */
	public boolean dependsOn(IMemoryGraphDelta delta) {
		boolean result = false;

		result = this.getAllDependentDeltas().contains(delta);

		return result;
	}

}
