/*
 * $Id: ReferenceTypeNode.java,v 1.6 2006/01/03 21:02:54 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 26.08.2004 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.type;

import java.util.Iterator;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.GraphElementUtil;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphElement;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.deltadebugging.ddstate.memorygraph.exceptions.MemoryGraphExtractionException;
import org.deltadebugging.ddstate.memorygraph.value.IValueNode;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaFieldVariable;
import org.eclipse.jdt.debug.core.IJavaReferenceType;

/**
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 * 
 * COMMENTME Comment type ***
 */
public class ReferenceTypeNode extends TypeNode implements IReferenceTypeNode {

    /** Personal logger for this class */
    private final static DDLogger logger = DDStatePlugin
            .getLogger(ReferenceTypeNode.class);

    /**
     * @param rootNode
     * @throws MemoryGraphExtractionException
     */
    public ReferenceTypeNode(IRootNode rootNode, IJavaReferenceType type)
            throws MemoryGraphExtractionException {
        super(rootNode, type);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.GraphNode#createSuccessors()
     */
    protected void createChildren() throws MemoryGraphExtractionException {
        // we ask for proxys to the fields of the reference type, walk through
        // them and create new nodes for every field we found
        IJavaReferenceType refType = (IJavaReferenceType) this.getType();

        try {
            String[] fieldNames = refType.getDeclaredFieldNames();

            for (int i = 0; i < fieldNames.length; i++) {
                String fieldName = fieldNames[i];
                IJavaFieldVariable field = refType.getField(fieldName);

                // if the field is not static the method getField returns null
                if (field != null) {
                    if (field.isStatic()) { // && !field.isFinal()) {
                        logger.trace(DDStatePlugin.DEBUG_GRAPH_EXTRACTOR,
                                "Create a new global successor of reference type '"
                                        + this.getTypeName() + "'\n"
                                        + "\t- field name = '" + fieldName
                                        + "'");

                        IValueNode successor = (IValueNode) this
                                .createAndRegisterNode(field);

                        IGraphEdge newSuccessor = GraphElementUtil
                                .createGlobalVariableEdge(this, successor,
                                        fieldName);

                        this.addSuccessor(newSuccessor);
                        successor.addParent(newSuccessor);
                    }
                }
            }

        } catch (DebugException e) {
            throw new MemoryGraphExtractionException(
                    "Problems on creating the successors of " + this.toString(),
                    e);
        }
    }

    public DeltaSet computeDeltaSet(IGraphElement element)
            throws DeltaComputationException {

        if (!this.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
            return new DeltaSet();
        }

        DeltaSet result = new DeltaSet();

        assert element instanceof IReferenceTypeNode;
        IReferenceTypeNode refType = (IReferenceTypeNode) element;

        assert this.isAssociatedWith(refType);
        assert this.getTypeName().equals(refType.getTypeName());

        // Both class type references has the same type, also they have the
        // same number of successors (static attributes) with the same names.
        // Because the successors are sorted by name, we can go through the
        // successors and compare each other.
        assert this.getChildren().size() == refType.getChildren().size();

        // Two graph node class references types are equal, if and only if they
        // have the same type name and are associated to each other.
        // Both criteria are preconditions of this method.
        GraphDecorator.decorateGraphElementsAsEqual(this, element);

        Iterator itFirstSuccessors = this.getChildren().iterator();
        Iterator itSecondSuccessors = refType.getChildren().iterator();

        IGraphEdge firstSuccessor = null;
        IGraphEdge secondSuccessor = null;

        while (itFirstSuccessors.hasNext()) {
            firstSuccessor = (IGraphEdge) itFirstSuccessors.next();
            secondSuccessor = (IGraphEdge) itSecondSuccessors.next();

            result.addAll(firstSuccessor.computeDeltaSet(secondSuccessor));
        }

        return result;
    }
}