/*
 * $Id: GraphNodePrimitiveArrayReference.java,v 1.14 2006/01/03 21:02:54 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 24.08.2005 23:33:29 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.value;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.memorygraph.GraphElementUtil;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphElement;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreatePrimitiveValueDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateArraySizeDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdatePrimitiveValueDelta;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.deltadebugging.ddstate.memorygraph.exceptions.MemoryGraphExtractionException;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaArray;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * This array reference represents primitive arrays. We introduced this special
 * type to speed up the graph extraction. This type constructs not for all his
 * successors (primitive values) new graph nodes. First at the graph comparation
 * if their exists differences between two primitive array references the
 * required successors will construct.
 * <p>
 * The result of this special type that we speed up the graph extraction and
 * saving a lot of memory space, because we don't construct the primitive graph
 * nodes for every array element.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
 */
public class GraphNodePrimitiveArrayReference extends ArrayNode {

    public GraphNodePrimitiveArrayReference(IRootNode rootNode, IJavaArray value)
            throws MemoryGraphExtractionException {
        super(rootNode, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#computeDeltaSet(org.deltadebugging.ddstate.memorygraph.IGraphElement)
     */
    public DeltaSet computeDeltaSet(IGraphElement element)
            throws DeltaComputationException {

        if (!this.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
            return new DeltaSet();
        }

        DeltaSet result = new DeltaSet();

        assert element instanceof GraphNodePrimitiveArrayReference;
        GraphNodePrimitiveArrayReference arrRef = (GraphNodePrimitiveArrayReference) element;

        assert this.isAssociatedWith(arrRef);
        assert this.getReferenceTypeName()
                .equals(arrRef.getReferenceTypeName());

        if (this.getLength() != arrRef.getLength()) {
            UpdateArraySizeDelta updateSizeDelta = new UpdateArraySizeDelta(
                    this, arrRef);
            result.add(updateSizeDelta);
            GraphDecorator.decorateGraphElementsAsDifferent(this, arrRef);

            int minIndex = Math.min(this.getLength(), arrRef.getLength());

            if (this.getLength() > arrRef.getLength()) {
                // This deltas are not realy necessary (outcomment 2005-09-18)
                // try {
                // result.addAll(this.handleDeletedElements(minIndex));
                // } catch (DebugException e) {
                // throw new DeltaComputationException(
                // "Failed to create the deltas "
                // + "for the deleted elements in the "
                // + "graph node primitive array reference", e);
                // } catch (MemoryGraphExtractionException e) {
                // throw new DeltaComputationException(
                // "Failed to create the deltas "
                // + "for the deleted elements in the "
                // + "graph node primitive array reference", e);
                // }
            } else {
                try {
                    result.addAll(this.handleAddedElements(arrRef, minIndex,
                            updateSizeDelta));
                } catch (DebugException e) {
                    throw new DeltaComputationException(
                            "Failed to create the deltas "
                                    + "for the added elements in the "
                                    + "graph node primitive array reference", e);
                } catch (MemoryGraphExtractionException e) {
                    throw new DeltaComputationException(
                            "Failed to create the deltas "
                                    + "for the added elements in the "
                                    + "graph node primitive array reference", e);
                }
            }

        } else {
            GraphDecorator.decorateGraphElementsAsEqual(this, arrRef);

        }

        // compare primitive values
        try {
            result.addAll(handleUpdatedElements(arrRef));
        } catch (DebugException e) {
            throw new DeltaComputationException("Failed to create the deltas "
                    + "for the updated elements in the "
                    + "graph node primitive array reference", e);
        } catch (MemoryGraphExtractionException e) {
            throw new DeltaComputationException("Failed to create the deltas "
                    + "for the updated elements in the "
                    + "graph node primitive array reference", e);
        }

        return result;
    }

    /**
     * Compute the update primitive value deltas between this array and the
     * given array.
     * 
     * @param arrRef
     * @return the delta set of all created update deltas.
     * @throws DebugException
     *             if the retriev of the variable for the computed index failed.
     * @throws MemoryGraphExtractionException
     *             if the creation of the new successor for the computed index
     *             failed.
     */
    private DeltaSet handleUpdatedElements(
            GraphNodePrimitiveArrayReference arrRef) throws DebugException,
            MemoryGraphExtractionException {
        DeltaSet result = new DeltaSet();

        IJavaArray thisArray = (IJavaArray) this.getValue();
        IJavaArray refArray = (IJavaArray) arrRef.getValue();

        int minIndex = Math.min(this.getLength(), arrRef.getLength());

        for (int i = 0; i < minIndex; i++) {
            if (!thisArray.getValue(i).getValueString().equals(
                    refArray.getValue(i).getValueString())) {

                // add successor of this array
                IPrimitiveValueNode oldValueNode = (IPrimitiveValueNode) GraphElementUtil
                        .createValueNode(this.getRootNode(),
                                (IJavaVariable) thisArray.getVariable(i));

                IGraphEdge oldValueEdge = GraphElementUtil
                        .createArrayElementEdge(this, oldValueNode, i);

                oldValueNode.addParent(oldValueEdge);
                this.addSuccessor(oldValueEdge);

                // add successor of other array
                IPrimitiveValueNode newValueNode = (IPrimitiveValueNode) GraphElementUtil
                        .createValueNode(arrRef.getRootNode(),
                                (IJavaVariable) refArray.getVariable(i));

                IGraphEdge newValueEdge = GraphElementUtil
                        .createArrayElementEdge(arrRef, newValueNode, i);

                newValueNode.addParent(newValueEdge);
                arrRef.addSuccessor(newValueEdge);

                // set associations
                oldValueNode.setAssociatedElement(newValueNode);
                newValueNode.setAssociatedElement(oldValueNode);

                oldValueEdge.setAssociatedElement(newValueEdge);
                newValueEdge.setAssociatedElement(oldValueEdge);

                // decorate
                GraphDecorator.decorateGraphElementsAsEqual(oldValueEdge,
                        newValueEdge);
                GraphDecorator.decorateGraphElementsAsDifferent(oldValueNode,
                        newValueNode);

                result.add(new UpdatePrimitiveValueDelta(oldValueNode,
                        newValueNode));

            }
        }

        return result;
    }

    /**
     * Creates create primitive value deltas for the given array beginning at
     * the given startIndex. Register the given update array size delta as
     * dependency of the new created deltas.
     * 
     * @param arrRef
     * @param startIndex
     * @param updateArraySizeDelta
     * @return the delta set of all created create deltas.
     * @throws DebugException
     *             if the retriev of the variable for the computed index failed.
     * @throws MemoryGraphExtractionException
     *             if the creation of the new successor for the computed index
     *             failed.
     */
    private DeltaSet handleAddedElements(
            GraphNodePrimitiveArrayReference arrRef, int startIndex,
            UpdateArraySizeDelta updateArraySizeDelta) throws DebugException,
            MemoryGraphExtractionException {
        DeltaSet result = new DeltaSet();

        for (int i = startIndex; i < arrRef.getLength(); i++) {
            IJavaVariable element = (IJavaVariable) ((IJavaArray) arrRef
                    .getValue()).getVariable(i);

            IPrimitiveValueNode newNode = (IPrimitiveValueNode) GraphElementUtil
                    .createValueNode(arrRef.getRootNode(), element);

            IGraphEdge newEdge = GraphElementUtil.createArrayElementEdge(
                    arrRef, newNode, i);

            newNode.addParent(newEdge);
            arrRef.addSuccessor(newEdge);

            GraphDecorator.decorateGraphEdgeAs(newEdge,
                    GraphDecorator.ONLY_IN_SECONDGRAPH, true);

            IMemoryGraphDelta newDelta = new CreatePrimitiveValueDelta(
                    updateArraySizeDelta, newEdge);

            result.add(newDelta);

        }

        return result;
    }

    // /**
    // * Creates delete primitive value deltas for this array beginning at the
    // * given startIndex.
    // *
    // * @param startIndex
    // * @return the delta set of all created delete deltas.
    // * @throws DebugException
    // * if the retriev of the variable for the computed index failed.
    // * @throws MemoryGraphExtractionException
    // * if the creation of the new successor for the computed index
    // * failed.
    // */
    // private DeltaSet handleDeletedElements(int startIndex)
    // throws DebugException, MemoryGraphExtractionException {
    // DeltaSet result = new DeltaSet();
    //
    // for (int i = startIndex; i < this.getLength(); i++) {
    // IJavaVariable element = (IJavaVariable) ((IJavaArray) this
    // .getValue()).getVariable(i);
    //
    // IPrimitiveValueNode newNode = (IPrimitiveValueNode)
    // GraphElementUtil
    // .createNode(this.getRootNode(), element, "[" + i + "]");
    //
    // IGraphEdge newEdge = GraphElementUtil.createArrayElementEdge(
    // this.getRootNode(), this, newNode, i);
    //
    // newNode.addPredeccessor(newEdge);
    // this.addSuccessor(newEdge);
    //
    // GraphDecorator.decorateGraphEdgeAs(newEdge,
    // GraphDecorator.ONLY_IN_FIRSTGRAPH, true);
    //
    // result.add(new DeletePrimitiveValueDelta(newNode));
    //
    // }
    //
    // return result;
    // }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.GraphNode#createSuccessors()
     */
    protected void createChildren() throws MemoryGraphExtractionException {
        // do nothing
    }

}
