/**
 * <p>Title: Bytecode instrumentation</p>
 * <p>Description: Tracing framework for java bytecode</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universitt Saarbrcken</p>
 * @author Valentin Dallmeier
 * @version 1.0
 */
package org.softevo.sibrelib.runtime.util;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import org.apache.bcel.generic.LocalVariableGen;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.Type;

/**
 * A collection of static utility methods used throughout octopus. <BR>
 */
public class Utilities {

    /**
     * This method tests if a file name points to a jar. <BR>
     * 
     * @param fileName
     * @return <code>true</code> if it is the name of a zip file,
     *         <code>false</code> otherwise
     */
    public static boolean isZipFileName(String fileName) {
        return isZipFile(new File(fileName));
    }

    /**
     * This method tests if a given file is a zip file or not. <BR>
     * 
     * @param candidate
     *            the file that is tested
     * @return <code>true</code> if the file is a zip file, <code>false</code>
     *         otherwise
     */
    public static boolean isZipFile(File candidate) {
        try {
            new ZipFile(candidate);
            return true;
        } catch (IOException exception) {
            return false;
        }
    }

    /**
     * This method reads the class path from the system properties and splits
     * them with the path separator from the system properties. <BR>
     * 
     * @return the tokenized class path
     */
    public static String[] getClassPathEntries() {
        return getClassPathEntries(System.getProperty("java.class.path"));
    }

    /**
     * This method splits the string passed with the path separator from the
     * system properties. <BR>
     * 
     * @return the tokenized class path
     * @param classPath
     *            the system class path
     */
    public static String[] getClassPathEntries(String classPath) {
        StringTokenizer splitter;
        String[] result;
        int counter = 0;

        splitter = new StringTokenizer(classPath, System
                .getProperty("path.separator"));
        result = new String[splitter.countTokens()];
        while (splitter.hasMoreTokens()) {
            result[counter] = splitter.nextToken();
            counter = counter + 1;
        }
        return result;
    }

    /**
     * This method gets a set of the names of all classes found in a zip file.
     * <BR>
     * 
     * @param zipFileName
     *            the name of the zip file
     * @return a set with the fully qualified class names of all classes in the
     *         zip
     * @throws IOException
     *             if an exception occurs processing the zip
     */
    public static HashMap getClassesInZip(String zipFileName)
            throws IOException {
        HashMap classesInZip;
        ZipFile zipFile;
        ZipEntry zipEntry;
        Enumeration entries;

        zipFile = new ZipFile(zipFileName);
        entries = zipFile.entries();
        classesInZip = new HashMap(zipFile.size());
        while (entries.hasMoreElements()) {
            zipEntry = (ZipEntry) entries.nextElement();
            if (zipEntry.getName().endsWith(".class")) {
                classesInZip.put(getClassNameForFileName(zipEntry.getName()),
                        zipEntry);
            }
        }
        return classesInZip;
    }

    /**
     * This method removes the base directory part of a file name. <BR>
     * 
     * @param file
     *            the file to remove the base part from
     * @param baseDirectory
     *            the base directory
     * @return the rest of the filename or the filename itself if it is not
     *         located in the base directory
     */
    public static String removeBaseDirectoryPart(File file, File baseDirectory) {
        String filePath, baseDirectoryPath;
        filePath = file.getAbsolutePath();
        baseDirectoryPath = baseDirectory.getAbsolutePath();
        if (filePath.length() > baseDirectoryPath.length()) {
            new File(filePath.substring(0, baseDirectoryPath.length()));
            /** compareTo does not seem to work* */
            //if (testFile.compareTo(baseDirectory) == 0) {
            return filePath.substring(baseDirectoryPath.length() + 1, filePath
                    .length());
            //}
        }
        return filePath;
    }

    /**
     * This method gets the file name that corresponds to the class name passed.
     * <BR>
     * All dots are replaced with slashes and ".class" is appended.
     * 
     * @param className
     *            the name of the class
     * @return the name of the file for the class
     */
    public static String getFileNameForClassName(String className) {
        return className.replace('.', '/').concat(".class");
    }

    /**
     * This method gets the class name that corresponds to the file name passed.
     * <BR>
     * All slashes are replaced with dots and the ".class" suffix is removed if
     * present. The file name is expected to be relative to the root directory
     * of the package hierachy.
     * 
     * @param fileName
     *            the name of the class
     * @return the name of the class
     */
    public static String getClassNameForFileName(String fileName) {
        if (fileName.endsWith(".class")) {
            fileName = fileName.substring(0, fileName.length() - 6);
        }
        return fileName.replace('/', '.');
    }

    /**
     * Prints the keys and entries of a <code>Map</code> to
     * <code>System.out</code>.
     * 
     * @param map
     *            the map that is displayed
     */
    public static void displayMap(Map map) {
        Iterator scanner;
        Object key;

        scanner = map.keySet().iterator();
        while (scanner.hasNext()) {
            key = scanner.next();
            System.out.println(key + " : " + map.get(key));
        }
    }

    /**
     * Prints the content of a <code>Collection</code> to
     * <code>System.out</code>.<BR>
     * 
     * @param collection
     *            the collection that is displayed
     */
    public static void displayCollection(Collection collection) {
        Iterator scanner;
        int counter = 0;

        scanner = collection.iterator();
        while (scanner.hasNext()) {
            System.out.println(counter + ". " + scanner.next());
            counter = counter + 1;
        }
    }

    /**
     * Deletes a whole directory and all files in all subdirectories. <BR>
     * Some files may not be deleted if the current user does not have write
     * access. This case is not reflected by an exception.
     * 
     * @param directory
     *            the directory to delete
     */
    public static void deleteDirectory(File directory) {
        File[] children;

        if (directory.exists()) {
            children = directory.listFiles();
            for (int counter = 0; counter < children.length; counter++) {
                if (children[counter].isDirectory()) {
                    deleteDirectory(children[counter]);
                } else {
                    children[counter].delete();
                }
            }
            directory.delete();
        }
    }

    /**
     * This method reads all bytes from an <code>InputStream</code> until it
     * blocks and returns them. <BR>
     * Note that this method does not close the stream after reading as this
     * should be done externally.
     * 
     * @param inputStream
     *            the stream to read data from
     * @return the bytes read
     * @throws IOException
     *             if an exception occurs reading from the stream
     */
    public static byte[] readAllBytes(InputStream inputStream)
            throws IOException {
        ByteArrayOutputStream byteBuffer;
        int byteRead;
        int bufferSize;

        bufferSize = inputStream.available();
        if (bufferSize <= 0) {
            bufferSize = 1000;
        }
        byteBuffer = new ByteArrayOutputStream();
        while ((byteRead = inputStream.read()) >= 0) {
            byteBuffer.write(byteRead);
        }
        return byteBuffer.toByteArray();
    }

    /**
     * Removes all occurences of a character from a string. <BR>
     * 
     * @param type
     *            the string the character is removed from
     * @return the string without the character
     */
    public static String removeCharacterFromString(String signature,
            char character) {
        String result = null;
        StringBuffer strBuf;

        if (signature.indexOf('[') < 0) {
            result = signature;

        } else {
            strBuf = new StringBuffer(signature.length());
            for (int counter = 0; counter < signature.length(); counter++) {
                if (signature.charAt(counter) != character) {
                    strBuf.append(signature.charAt(counter));
                }
            }

            result = strBuf.toString();
        }

        return result;
    }

    /**
     * Gets the index for the next local variable to append in a method. <BR>
     * This is in order to avoid a verify error (register pair) whenever at
     * least one local variable is a double or long.
     * 
     * @param methodGenerator
     *            the generator for the method
     * @return int the index that may be used for the next method
     */
    public static int getNextVariableIndex(MethodGen methodGenerator) {
        int result = 0;

        LocalVariableGen lastVariableGen;

        if (methodGenerator.getLocalVariables().length > 0) {
            lastVariableGen = methodGenerator.getLocalVariables()[methodGenerator
                    .getLocalVariables().length - 1];
            if (lastVariableGen.getType().equals(Type.DOUBLE)
                    || lastVariableGen.getType().equals(Type.LONG)) {
                result = lastVariableGen.getIndex() + 2;
            } else {
                result = lastVariableGen.getIndex() + 1;
            }
        } else {
            result = 0;
        }

        return result;
    }

}
