/**
 * <p>
 * Title: Bytecode instrumentation
 * </p>
 * <p>
 * Description: Tracing framework for java bytecode
 * </p>
 * <p>
 * Copyright: Copyright (c) 2004
 * </p>
 * <p>
 * Company: Universitt Saarbrcken
 * </p>
 * 
 * @author Valentin Dallmeier
 * @version 1.0
 */
package org.softevo.sibrelib.source;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import org.softevo.sibrelib.runtime.util.Utilities;

public class JarSource extends ClassFileSource {

    /**
     * The source for classes (either jar or zip). <BR>
     */
    private ZipFile zipFile = null;

    /**
     * Maps class names to zip entries for all classes in the jar. <BR>
     */
    private HashMap classesInZip = null;

    /**
     * The target for save operations. <BR>
     */
    private File destination = null;

    /**
     * Creates a new class source that reads from a jar or zip file. <BR>
     * 
     * @param zipFile
     *            the zip
     * @param destination
     *            the target for modified content
     * @throws IOException
     *             if an exception occurs analysing the zip
     */
    public JarSource(ZipFile zipFile, File destination) throws IOException {
        this.zipFile = zipFile;
        this.destination = destination;
        this.classesInZip = Utilities.getClassesInZip(zipFile.getName());
    }

    /**
     * Saves the modified version of the jar file. <BR>
     * 
     * @throws IOException
     *             if an exception occurs writing the result file
     */
    public void saveModifiedVersion() throws IOException {
        ZipOutputStream zipOutputStream;
        Enumeration entries;
        ZipEntry entry;
        ClassFile classFile;
        byte[] bytes;
        OutputStream outputStream = null;

        try {
            outputStream = new FileOutputStream(this.destination);
            zipOutputStream = new ZipOutputStream(outputStream);
            entries = this.zipFile.entries();
            while (entries.hasMoreElements()) {
                entry = (ZipEntry) entries.nextElement();
                if (entry.getName().endsWith(".class")) {
                    classFile = (ClassFile) this.classesCreated.get(Utilities
                            .getClassNameForFileName(entry.getName()));
                    if (classFile != null) {
                        if (classFile.isModified()) {
                            bytes = classFile.getModifiedClassBytes();
                        } else {
                            bytes = classFile.getClassBytes();
                        }
                    } else {
                        bytes = Utilities.readAllBytes(this.zipFile
                                .getInputStream(entry));
                    }
                } else {
                    bytes = Utilities.readAllBytes(this.zipFile
                            .getInputStream(entry));
                }
                zipOutputStream.putNextEntry(new ZipEntry(entry.getName()));
                zipOutputStream.write(bytes);
            }
            zipOutputStream.finish();
        } finally {
            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException finalizeException) {
                    //ignore exception
                }
            }
        }
    }

    /**
     * Gets the class path that is needed in order to load all classes from the
     * source. <BR>
     */
    public String getClassPath() {
        return this.zipFile.getName();
    }

    /**
     * This class uses an <code>URLClassLoader</code> for the zip or jar file.
     * <BR>
     * 
     * @return the class loader for this zip or <code>null</code> if the URL
     *         can not be created (which should never occur)
     */
    public ClassLoader createBCELClassLoader() {
        return new BCELClassLoader(this);
    }

    /**
     * This method converts the class name passed into the corresponding file
     * name and tries to load the file from the zip. <BR>
     * 
     * @param className
     *            the name of the class sought
     * @return the <code>ClassFile</code> loaded
     * @throws NoSuchElementException
     *             if no class with the given name was found
     * @throws IOException
     *             if an exception occurs reading from the zip file
     */
    protected ClassFile findClassFile(String className)
            throws NoSuchElementException, IOException {
        ZipEntry entry;

        if (this.classesInZip.containsKey(className)) {
            Utilities.getFileNameForClassName(className);
            entry = (ZipEntry) this.classesInZip.get(className);
            return new ClassFile(className, Utilities.readAllBytes(this.zipFile
                    .getInputStream((entry))));
        }
        // else
        throw new NoSuchElementException("Zip file " + this.zipFile.getName()
                + " does not contain class " + className);
    }

    /**
     * Gets an iterator for all <code>ClassFiles</code> in this source. <BR>
     * 
     * @return the iterator
     */
    public Iterator getClassIterator() {
        return new ClassFileIterator(this.classesInZip.keySet().iterator());
    }

    /**
     * Return the number of all <code>ClassFiles</code> in this source.
     * 
     * @return number of all <code>ClassFiles</code>.
     */
    public int getSize() {
        int result = 0;
        if (this.classesInZip != null) {
            result = this.classesInZip.size();
        }

        return result;
    }

}